
# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


r"""Implement a MSM class that builds a Markov state models from
microstate trajectories automatically computes important properties
and provides them for later access.

.. moduleauthor:: F. Noe <frank DOT noe AT fu-berlin DOT de>

"""


import numpy as _np


from pyemma.msm.models.msm import MSM as _MSM
from pyemma.util import types as _types
from pyemma.util.annotators import fix_docs


@fix_docs
class HMSM(_MSM):
    r""" Hidden Markov model on discrete states.

    Parameters
    ----------
    hmm : :class:`DiscreteHMM <bhmm.DiscreteHMM>`
        Hidden Markov Model

    """
    __serialize_version = 0

    def __init__(self, P, pobs, pi=None, dt_model='1 step'):
        """

        Parameters
        ----------
        Pcoarse : ndarray (m,m)
            coarse-grained or hidden transition matrix

        Pobs : ndarray (m,n)
            observation probability matrix from hidden to observable discrete states

        dt_model : str, optional, default='1 step'
            time step of the model

        """
        # construct superclass and check input
        _MSM.__init__(self, P, dt_model=dt_model)
        self.set_model_params(P=P, pobs=pobs, pi=pi, dt_model=dt_model)

    def set_model_params(self, P=None, pobs=None, pi=None, reversible=None, dt_model='1 step', neig=None):
        """

        Parameters
        ----------
        P : ndarray(m,m)
            coarse-grained or hidden transition matrix

        pobs : ndarray (m,n)
            observation probability matrix from hidden to observable discrete
            states

        pi : ndarray(m), optional, default=None
            stationary or distribution. Can be optionally given in case if
            it was already computed, e.g. by the estimator.

        reversible : bool, optional, default=None
            whether P is reversible with respect to its stationary distribution.
            If None (default), will be determined from P

        dt_model : str, optional, default='1 step'
            time step of the model

        neig : int or None
            The number of eigenvalues / eigenvectors to be kept. If set to
            None, all eigenvalues will be used.

        """
        _MSM.set_model_params(self, P=P, pi=pi, reversible=reversible, dt_model=dt_model, neig=neig)
        # set P and derived quantities if available
        if pobs is not None:
            # check and save copy of output probability
            assert _types.is_float_matrix(pobs), 'pobs is not a matrix of floating numbers'
            assert _np.allclose(pobs.sum(axis=1), 1), 'pobs is not a stochastic matrix'
            self._nstates_obs = pobs.shape[1]
            self.update_model_params(pobs=pobs)

    @property
    def observation_probabilities(self):
        r""" returns the output probability matrix

        Returns
        -------
        Pout : ndarray (m,n)
            output probability matrix from hidden to observable discrete states

        """
        return self.pobs

    @property
    def nstates_obs(self):
        return self.observation_probabilities.shape[1]

    @property
    def lifetimes(self):
        r""" Lifetimes of states of the hidden transition matrix

        Returns
        -------
        l : ndarray(nstates)
            state lifetimes in units of the input trajectory time step,
            defined by :math:`-\tau / ln \mid p_{ii} \mid, i = 1,...,nstates`, where
            :math:`p_{ii}` are the diagonal entries of the hidden transition matrix.

        """
        return -self._timeunit_model.dt / _np.log(_np.diag(self.transition_matrix))

    def transition_matrix_obs(self, k=1):
        r""" Computes the transition matrix between observed states

        Transition matrices for longer lag times than the one used to
        parametrize this HMSM can be obtained by setting the k option.
        Note that a HMSM is not Markovian, thus we cannot compute transition
        matrices at longer lag times using the Chapman-Kolmogorow equality.
        I.e.:

        .. math::
            P (k \tau) \neq P^k (\tau)

        This function computes the correct transition matrix using the
        metastable (coarse) transition matrix :math:`P_c` as:

        .. math::
            P (k \tau) = {\Pi}^-1 \chi^{\top} ({\Pi}_c) P_c^k (\tau) \chi

        where :math:`\chi` is the output probability matrix, :math:`\Pi_c` is
        a diagonal matrix with the metastable-state (coarse) stationary
        distribution and :math:`\Pi` is a diagonal matrix with the
        observable-state stationary distribution.

        Parameters
        ----------
        k : int, optional, default=1
            Multiple of the lag time for which the
            By default (k=1), the transition matrix at the lag time used to
            construct this HMSM will be returned. If a higher power is given,

        """
        Pi_c = _np.diag(self.stationary_distribution)
        P_c = self.transition_matrix
        P_c_k = _np.linalg.matrix_power(P_c, k)  # take a power if needed
        B = self.observation_probabilities
        C = _np.dot(_np.dot(B.T, Pi_c), _np.dot(P_c_k, B))
        P = C / C.sum(axis=1)[:, None]  # row normalization
        return P

    @property
    def stationary_distribution_obs(self):
        return _np.dot(self.stationary_distribution, self.observation_probabilities)

    @property
    def eigenvectors_left_obs(self):
        return _np.dot(self.eigenvectors_left(), self.observation_probabilities)

    @property
    def eigenvectors_right_obs(self):
        return _np.dot(self.metastable_memberships, self.eigenvectors_right())

    def propagate(self, p0, k):
        r""" Propagates the initial distribution p0 k times

        Computes the product

        .. math::

            p_k = p_0^T P^k

        If the lag time of transition matrix :math:`P` is :math:`\tau`, this
        will provide the probability distribution at time :math:`k \tau`.

        Parameters
        ----------
        p0 : ndarray(n)
            Initial distribution. Vector of size of the active set.

        k : int
            Number of time steps

        Returns
        ----------
        pk : ndarray(n)
            Distribution after k steps. Vector of size of the active set.

        """
        p0 = _types.ensure_ndarray(p0, ndim=1, kind='numeric')
        assert _types.is_int(k) and k>=0, 'k must be a non-negative integer'
        if k == 0:  # simply return p0 normalized
            return p0 / p0.sum()

        micro = False
        # are we on microstates space?
        if len(p0) == self.nstates_obs:
            micro = True
            # project to hidden and compute
            p0 = _np.dot(self.observation_probabilities, p0)

        self._ensure_eigendecomposition(self.nstates)
        from pyemma.util.linalg import mdot
        pk = mdot(p0.T, self.eigenvectors_right(), _np.diag(_np.power(self.eigenvalues(), k)), self.eigenvectors_left())

        if micro:
            pk = _np.dot(pk, self.observation_probabilities)  # convert back to microstate space

        # normalize to 1.0 and return
        return pk / pk.sum()

    def submodel(self, states=None, obs=None):
        """Returns a HMM with restricted state space

        Parameters
        ----------
        states : None or int-array
            Hidden states to restrict the model to (if not None).
        obs : None, str or int-array
            Observed states to restrict the model to (if not None).

        Returns
        -------
        hmm : HMM
            The restricted HMM.

        """
        if states is None and obs is None:
            return self  # do nothing
        if states is None:
            states = _np.arange(self.nstates)
        if obs is None:
            obs = _np.arange(self.nstates_obs)

        # transition matrix
        P = self.transition_matrix[_np.ix_(states, states)].copy()
        P /= P.sum(axis=1)[:, None]

        # observation matrix
        B = self.observation_probabilities[_np.ix_(states, obs)].copy()
        B /= B.sum(axis=1)[:, None]

        sub_hmsm = HMSM(P, B, dt_model=self.dt_model)
        sub_hmsm.update_model_params(reversible=self.reversible)
        return sub_hmsm


    # ================================================================================================================
    # Experimental properties: Here we allow to use either coarse-grained or microstate observables
    # ================================================================================================================

    def expectation(self, a):
        a = _types.ensure_float_vector(a, require_order=True)
        # are we on microstates space?
        if len(a) == self.nstates_obs:
            # project to hidden and compute
            a = _np.dot(self.observation_probabilities, a)
        # now we are on macrostate space, or something is wrong
        if len(a) == self.nstates:
            return _MSM.expectation(self, a)
        else:
            raise ValueError('observable vector a has size %s which is incompatible with both hidden (%s) '
                             'and observed states (%s)' % (len(a), self.nstates, self.nstates_obs))

    def correlation(self, a, b=None, maxtime=None, k=None, ncv=None):
        # basic checks for a and b
        a = _types.ensure_ndarray(a, ndim=1, kind='numeric')
        b = _types.ensure_ndarray_or_None(b, ndim=1, kind='numeric', size=len(a))
        # are we on microstates space?
        if len(a) == self.nstates_obs:
            a = _np.dot(self.observation_probabilities, a)
            if b is not None:
                b = _np.dot(self.observation_probabilities, b)
        # now we are on macrostate space, or something is wrong
        if len(a) == self.nstates:
            return _MSM.correlation(self, a, b=b, maxtime=maxtime)
        else:
            raise ValueError('observable vectors have size %s which is incompatible with both hidden (%s)'
                             ' and observed states (%s)' % (len(a), self.nstates, self.nstates_obs))

    def fingerprint_correlation(self, a, b=None, k=None, ncv=None):
        # basic checks for a and b
        a = _types.ensure_ndarray(a, ndim=1, kind='numeric')
        b = _types.ensure_ndarray_or_None(b, ndim=1, kind='numeric', size=len(a))
        # are we on microstates space?
        if len(a) == self.nstates_obs:
            a = _np.dot(self.observation_probabilities, a)
            if b is not None:
                b = _np.dot(self.observation_probabilities, b)
        # now we are on macrostate space, or something is wrong
        if len(a) == self.nstates:
            return _MSM.fingerprint_correlation(self, a, b=b)
        else:
            raise ValueError('observable vectors have size %s which is incompatible with both hidden (%s)'
                             ' and observed states (%s)' % (len(a), self.nstates, self.nstates_obs))

    def relaxation(self, p0, a, maxtime=None, k=None, ncv=None):
        # basic checks for a and b
        p0 = _types.ensure_ndarray(p0, ndim=1, kind='numeric')
        a = _types.ensure_ndarray(a, ndim=1, kind='numeric', size=len(p0))
        # are we on microstates space?
        if len(a) == self.nstates_obs:
            p0 = _np.dot(self.observation_probabilities, p0)
            a = _np.dot(self.observation_probabilities, a)
        # now we are on macrostate space, or something is wrong
        if len(a) == self.nstates:
            return _MSM.relaxation(self, p0, a, maxtime=maxtime)
        else:
            raise ValueError('observable vectors have size %s which is incompatible with both hidden (%s)'
                             ' and observed states (%s)' % (len(a), self.nstates, self.nstates_obs))

    def fingerprint_relaxation(self, p0, a, k=None, ncv=None):
        # basic checks for a and b
        p0 = _types.ensure_ndarray(p0, ndim=1, kind='numeric')
        a = _types.ensure_ndarray(a, ndim=1, kind='numeric', size=len(p0))
        # are we on microstates space?
        if len(a) == self.nstates_obs:
            p0 = _np.dot(self.observation_probabilities, p0)
            a = _np.dot(self.observation_probabilities, a)
        # now we are on macrostate space, or something is wrong
        if len(a) == self.nstates:
            return _MSM.fingerprint_relaxation(self, p0, a)
        else:
            raise ValueError('observable vectors have size %s which is incompatible with both hidden (%s)'
                             ' and observed states (%s)' % (len(a), self.nstates, self.nstates_obs))

    def pcca(self, m):
        raise NotImplementedError('PCCA is not meaningful for Hidden Markov models. ' +
                                  'If you really want to do this, initialize an MSM with the HMSM transition matrix.')

    # ================================================================================================================
    # Metastable state stuff is overwritten, because we now have the HMM output probability matrix
    # ================================================================================================================

    @property
    def metastable_memberships(self):
        r""" Computes the memberships of observable states to metastable sets by
            Bayesian inversion as described in [1]_.

        Returns
        -------
        M : ndarray((n,m))
            A matrix containing the probability or membership of each
            observable state to be assigned to each metastable or hidden state.
            The row sums of M are 1.

        References
        ----------
        .. [1] F. Noe, H. Wu, J.-H. Prinz and N. Plattner: Projected and hidden
            Markov models for calculating kinetics and metastable states of
            complex molecules. J. Chem. Phys. 139, 184114 (2013)

        """
        nonzero = _np.nonzero(self.stationary_distribution_obs)[0]
        M = _np.zeros((self.nstates_obs, self.nstates))
        M[nonzero, :] = _np.transpose(_np.diag(self.stationary_distribution).dot(
            self.observation_probabilities[:, nonzero]) / self.stationary_distribution_obs[nonzero])
        # renormalize
        M[nonzero, :] /= M.sum(axis=1)[nonzero, None]
        return M

    @property
    def metastable_distributions(self):
        r""" Returns the output probability distributions. Identical to
            :meth:`observation_probabilities`

        Returns
        -------
        Pout : ndarray (m,n)
            output probability matrix from hidden to observable discrete states

        See also
        --------
        observation_probabilities

        """
        return self.observation_probabilities

    @property
    def metastable_sets(self):
        r""" Computes the metastable sets of observable states within each
            metastable set

        Notes
        -----
        This is only recommended for visualization purposes. You *cannot*
        compute any actual quantity of the coarse-grained kinetics without
        employing the fuzzy memberships!

        Returns
        -------
        sets : list of int-arrays
            A list of length equal to metastable states. Each element is an array
            with observable state indexes contained in it

        """
        res = []
        assignment = self.metastable_assignments
        for i in range(self.nstates):
            res.append(_np.where(assignment == i)[0])
        return res

    @property
    def metastable_assignments(self):
        r""" Computes the assignment to metastable sets for observable states

        Notes
        -----
        This is only recommended for visualization purposes. You *cannot*
        compute any actual quantity of the coarse-grained kinetics without
        employing the fuzzy memberships!

        Returns
        -------
        ndarray((n) ,dtype=int)
            For each observable state, the metastable state it is located in.

        """
        return _np.argmax(self.observation_probabilities, axis=0)


    def simulate(self, N, start=None, stop=None, dt=1):
        """
        Generates a realization of the Hidden Markov Model

        Parameters
        ----------
        N : int
            trajectory length in steps of the lag time
        start : int, optional, default = None
            starting hidden state. If not given, will sample from the stationary
            distribution of the hidden transition matrix.
        stop : int or int-array-like, optional, default = None
            stopping hidden set. If given, the trajectory will be stopped before
            N steps once a hidden state of the stop set is reached
        dt : int
            trajectory will be saved every dt time steps.
            Internally, the dt'th power of P is taken to ensure a more efficient simulation.

        Returns
        -------
        htraj : (N/dt, ) ndarray
            The hidden state trajectory with length N/dt
        otraj : (N/dt, ) ndarray
            The observable state discrete trajectory with length N/dt

        """

        from scipy import stats
        import msmtools.generation as msmgen
        # generate output distributions
        output_distributions = [stats.rv_discrete(values=(_np.arange(self.pobs.shape[1]), pobs_i)) for pobs_i in self.pobs]
        # sample hidden trajectory
        htraj = msmgen.generate_traj(self.transition_matrix, N, start=start, stop=stop, dt=dt)
        otraj = _np.zeros(htraj.size, dtype=int)
        # for each time step, sample microstate
        for t, h in enumerate(htraj):
            otraj[t] = output_distributions[h].rvs()  # current cluster
        return htraj, otraj



# This file is part of PyEMMA.
#
# Copyright (c) 2016-2017 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as _np
import warnings as _warnings

from pyemma._base.estimator import Estimator as _Estimator
from pyemma._base.progress import ProgressReporter as _ProgressReporter
from pyemma.thermo import MEMM as _MEMM
from pyemma.thermo.estimators._base import ThermoBase
from pyemma.thermo.models.memm import ThermoMSM as _ThermoMSM
from pyemma.util import types as _types
from pyemma.thermo.estimators._callback import _ConvergenceProgressIndicatorCallBack
from pyemma.thermo.estimators._callback import _IterationProgressIndicatorCallBack


from pyemma.thermo.extensions import (
    tram as _tram,
    tram_direct as _tram_direct,
    trammbar as _trammbar,
    trammbar_direct as _trammbar_direct,
    mbar as _mbar,
    mbar_direct as _mbar_direct,
    util as _util,
    cset as _cset,
)

from msmtools.estimation import largest_connected_set as _largest_connected_set


class EmptyState(RuntimeWarning):
    pass


class TRAM(_Estimator, _MEMM, ThermoBase):
    r"""Transition(-based) Reweighting Analysis Method."""
    __serialize_version = 0
    __serialize_fields = ('biased_conf_energies',
                          'btrajs',
                          'count_matrices',
                          'csets',
                          'dtrajs',
                          'equilibrium_btrajs',
                          'equilibrium_dtrajs',
                          'equilibrium_state_counts',
                          'equilibrium_state_counts_full',
                          'increments',
                          'log_lagrangian_mult',
                          'loglikelihoods',
                          'mbar_biased_conf_energies',
                          'mbar_therm_energies',
                          'mbar_unbiased_conf_energies',
                          'nthermo',
                          'state_counts',
                          'therm_energies',
                          'therm_state_counts_full',
                          )

    def __init__(
        self, lag, count_mode='sliding',
        connectivity='post_hoc_RE',
        nstates_full=None, equilibrium=None,
        maxiter=10000, maxerr=1.0E-15, save_convergence_info=0, dt_traj='1 step',
        nn=None, connectivity_factor=1.0, direct_space=False, N_dtram_accelerations=0,
        callback=None,
        init='mbar', init_maxiter=5000, init_maxerr=1.0E-8,
        overcounting_factor=1.0):
        r"""Transition(-based) Reweighting Analysis Method

        Parameters
        ----------
        lag : int
            Integer lag time at which transitions are counted.
        count_mode : str, optional, default='sliding'
            mode to obtain count matrices from discrete trajectories. Should be
            one of:
            * 'sliding' : A trajectory of length T will have :math:`T-\tau` counts at time indexes
                  .. math::
                     (0 \rightarrow \tau), (1 \rightarrow \tau+1), ..., (T-\tau-1 \rightarrow T-1)
            * 'sample' : A trajectory of length T will have :math:`T/\tau` counts
              at time indexes
                  .. math::
                        (0 \rightarrow \tau), (\tau \rightarrow 2 \tau), ..., ((T/\tau-1) \tau \rightarrow T)
            Currently only 'sliding' is supported.
        connectivity : str, optional, default='post_hoc_RE'
            One of 'post_hoc_RE', 'BAR_variance', 'reversible_pathways' or
            'summed_count_matrix'. Defines what should be considered a connected set
            in the joint (product) space of conformations and thermodynamic ensembles.
            * 'reversible_pathways' : requires that every state in the connected set
              can be reached by following a pathway of reversible transitions. A
              reversible transition between two Markov states (within the same
              thermodynamic state k) is a pair of Markov states that belong to the
              same strongly connected component of the count matrix (from
              thermodynamic state k). A pathway of reversible transitions is a list of
              reversible transitions [(i_1, i_2), (i_2, i_3),..., (i_(N-2), i_(N-1)),
              (i_(N-1), i_N)]. The thermodynamic state where the reversible
              transitions happen, is ignored in constructing the reversible pathways.
              This is equivalent to assuming that two ensembles overlap at some Markov
              state whenever there exist frames from both ensembles in that Markov
              state.
            * 'post_hoc_RE' : similar to 'reversible_pathways' but with a more strict
              requirement for the overlap between thermodynamic states. It is required
              that every state in the connected set can be reached by following a
              pathway of reversible transitions or jumping between overlapping
              thermodynamic states while staying in the same Markov state. A reversible
              transition between two Markov states (within the same thermodynamic
              state k) is a pair of Markov states that belong to the same strongly
              connected component of the count matrix (from thermodynamic state k).
              Two thermodynamic states k and l are defined to overlap at Markov state
              n if a replica exchange simulation [2]_ restricted to state n would show
              at least one transition from k to l or one transition from from l to k.
              The expected number of replica exchanges is estimated from the
              simulation data. The minimal number required of replica exchanges
              per Markov state can be increased by decreasing `connectivity_factor`.
            * 'BAR_variance' : like 'post_hoc_RE' but with a different condition to
              define the thermodynamic overlap based on the variance of the BAR
              estimator [3]_. Two thermodynamic states k and l are defined to overlap
              at Markov state n if the variance of the free energy difference Delta
              f_{kl} computed with BAR (and restricted to conformations form Markov
              state n) is less or equal than one. The minimally required variance
              can be controlled with `connectivity_factor`.
            * 'summed_count_matrix' : all thermodynamic states are assumed to overlap.
              The connected set is then computed by summing the count matrices over
              all thermodynamic states and taking it's largest strongly connected set.
              Not recommended!
            For more details see :func:`pyemma.thermo.extensions.cset.compute_csets_TRAM`.
        nstates_full : int, optional, default=None
            Number of cluster centers, i.e., the size of the full set of states.
        equilibrium : list of booleans, optional
            For every trajectory triple (ttraj[i], dtraj[i], btraj[i]), indicates
            whether to assume global equilibrium. If true, the triple is not used
            for computing kinetic quantities (but only thermodynamic quantities).
            By default, no trajectory is assumed to be in global equilibrium.
            This is the TRAMMBAR extension.
        maxiter : int, optional, default=10000
            The maximum number of self-consistent iterations before the estimator exits unsuccessfully.
        maxerr : float, optional, default=1E-15
            Convergence criterion based on the maximal free energy change in a self-consistent
            iteration step.
        save_convergence_info : int, optional, default=0
            Every save_convergence_info iteration steps, store the actual increment
            and the actual log-likelihood; 0 means no storage.
        dt_traj : str, optional, default='1 step'
            Description of the physical time corresponding to the lag. May be used by analysis
            algorithms such as plotting tools to pretty-print the axes. By default '1 step', i.e.
            there is no physical time unit.  Specify by a number, whitespace and unit. Permitted
            units are (* is an arbitrary string):

            |  'fs',   'femtosecond*'
            |  'ps',   'picosecond*'
            |  'ns',   'nanosecond*'
            |  'us',   'microsecond*'
            |  'ms',   'millisecond*'
            |  's',    'second*'
        connectivity_factor : float, optional, default=1.0
            Only needed if connectivity='post_hoc_RE' or 'BAR_variance'. Values
            greater than 1.0 weaken the connectivity conditions. For 'post_hoc_RE'
            this multiplies the number of hypothetically observed transitions. For
            'BAR_variance' this scales the threshold for the minimal allowed variance
            of free energy differences.
        direct_space : bool, optional, default=False
            Whether to perform the self-consistent iteration with Boltzmann factors
            (direct space) or free energies (log-space). When analyzing data from
            multi-temperature simulations, direct-space is not recommended.
        N_dtram_accelerations : int, optional, default=0
            Convergence of TRAM can be speeded up by interleaving the updates
            in the self-consistent iteration with a dTRAM-like update step.
            N_dtram_accelerations says how many times the dTRAM-like update
            step should be applied in every iteration of the TRAM equations.
            Currently this is only effective if direct_space=True.
        init : str, optional, default=None
            Use a specific initialization for self-consistent iteration:

            | None:    use a hard-coded guess for free energies and Lagrangian multipliers
            | 'mbar':  perform a short MBAR estimate to initialize the free energies
        init_maxiter : int, optional, default=5000
            The maximum number of self-consistent iterations during the initialization.
        init_maxerr : float, optional, default=1.0E-8
            Convergence criterion for the initialization.
        overcounting_factor : double, default = 1.0
            Only needed if equilibrium contains True (TRAMMBAR).
            Sets the relative statistical weight of equilibrium and non-equilibrium
            frames. An overcounting_factor of value n means that every
            non-equilibrium frame is counted n times. Values larger than 1 increase
            the relative weight of the non-equilibrium data. Values less than 1
            increase the relative weight of the equilibrium data.


        References
        ----------
        .. [1] Wu, H. et al 2016
            Multiensemble Markov models of molecular thermodynamics and kinetics
            Proc. Natl. Acad. Sci. USA 113 E3221--E3230
        .. [2]_ Hukushima et al, Exchange Monte Carlo method and application to spin
            glass simulations, J. Phys. Soc. Jan. 65, 1604 (1996)
        .. [3]_ Shirts and Chodera, Statistically optimal analysis of samples
            from multiple equilibrium states, J. Chem. Phys. 129, 124105 (2008)

        """
        self.lag = lag
        assert count_mode == 'sliding', 'Currently the only implemented count_mode is \'sliding\''
        self.count_mode = count_mode
        self.connectivity = connectivity
        self.nn = nn
        self.connectivity_factor = connectivity_factor
        self.dt_traj = dt_traj
        self.nstates_full = nstates_full
        self.equilibrium = equilibrium
        self.maxiter = maxiter
        self.maxerr = maxerr
        self.direct_space = direct_space
        self.N_dtram_accelerations = N_dtram_accelerations
        self.callback = callback
        self.save_convergence_info = save_convergence_info
        assert init in (None, 'mbar'), 'Currently only None and \'mbar\' are supported'
        self.init = init
        self.init_maxiter = init_maxiter
        self.init_maxerr = init_maxerr
        self.overcounting_factor = overcounting_factor
        self.active_set = None
        self.biased_conf_energies = None
        self.mbar_therm_energies = None
        self.log_lagrangian_mult = None
        self.loglikelihoods = None

    def estimate(self, X, **params):
        """
        Parameters
        ----------
        X : tuple of (ttrajs, dtrajs, btrajs)
            Simulation trajectories. ttrajs contain the indices of the thermodynamic state, dtrajs
            contains the indices of the configurational states and btrajs contain the biases.

            ttrajs : list of numpy.ndarray(X_i, dtype=int)
                Every element is a trajectory (time series). ttrajs[i][t] is the index of the
                thermodynamic state visited in trajectory i at time step t.
            dtrajs : list of numpy.ndarray(X_i, dtype=int)
                dtrajs[i][t] is the index of the configurational state (Markov state) visited in
                trajectory i at time step t.
            btrajs : list of numpy.ndarray((X_i, T), dtype=numpy.float64)
                For every simulation frame seen in trajectory i and time step t, btrajs[i][t,k] is the
                bias energy of that frame evaluated in the k'th thermodynamic state (i.e. at the k'th
                Umbrella/Hamiltonian/temperature).
        """
        return super(TRAM, self).estimate(X, **params)

    def _estimate(self, X):
        ttrajs, dtrajs_full, btrajs = X
        # shape and type checks
        assert len(ttrajs) == len(dtrajs_full) == len(btrajs)
        for t in ttrajs:
            _types.assert_array(t, ndim=1, kind='i')
        for d in dtrajs_full:
            _types.assert_array(d, ndim=1, kind='i')
        for b in btrajs:
            _types.assert_array(b, ndim=2, kind='f')
        # find dimensions
        nstates_full = max(_np.max(d) for d in dtrajs_full) + 1
        if self.nstates_full is None:
            self.nstates_full = nstates_full
        elif self.nstates_full < nstates_full:
            raise RuntimeError("Found more states (%d) than specified by nstates_full (%d)" % (
                nstates_full, self.nstates_full))
        self.nthermo = max(_np.max(t) for t in ttrajs) + 1
        # dimensionality checks
        for t, d, b, in zip(ttrajs, dtrajs_full, btrajs):
            assert t.shape[0] == d.shape[0] == b.shape[0]
            assert b.shape[1] == self.nthermo

        # cast types and change axis order if needed
        ttrajs = [_np.require(t, dtype=_np.intc, requirements='C') for t in ttrajs]
        dtrajs_full = [_np.require(d, dtype=_np.intc, requirements='C') for d in dtrajs_full]
        btrajs = [_np.require(b, dtype=_np.float64, requirements='C') for b in btrajs]

        # if equilibrium information is given, separate the trajectories
        if self.equilibrium is not None:
            assert len(self.equilibrium) == len(ttrajs)
            _ttrajs, _dtrajs_full, _btrajs = ttrajs, dtrajs_full, btrajs
            ttrajs = [ttraj for eq, ttraj in zip(self.equilibrium, _ttrajs) if not eq]
            dtrajs_full = [dtraj for eq, dtraj in zip(self.equilibrium, _dtrajs_full) if not eq]
            self.btrajs = [btraj for eq, btraj in zip(self.equilibrium, _btrajs) if not eq]
            equilibrium_ttrajs = [ttraj for eq, ttraj in zip(self.equilibrium, _ttrajs) if eq]
            equilibrium_dtrajs_full = [dtraj for eq, dtraj in zip(self.equilibrium, _dtrajs_full) if eq]
            self.equilibrium_btrajs = [btraj for eq, btraj in zip(self.equilibrium, _btrajs) if eq]
        else: # set dummy values
            equilibrium_ttrajs = []
            equilibrium_dtrajs_full = []
            self.equilibrium_btrajs = []
            self.btrajs = btrajs

        # find state visits and transition counts
        state_counts_full = _util.state_counts(ttrajs, dtrajs_full, nstates=self.nstates_full, nthermo=self.nthermo)
        count_matrices_full = _util.count_matrices(ttrajs, dtrajs_full,
            self.lag, sliding=self.count_mode, sparse_return=False, nstates=self.nstates_full, nthermo=self.nthermo)
        self.therm_state_counts_full = state_counts_full.sum(axis=1)

        if self.equilibrium is not None:
            self.equilibrium_state_counts_full = _util.state_counts(equilibrium_ttrajs, equilibrium_dtrajs_full,
                nstates=self.nstates_full, nthermo=self.nthermo)
        else:
            self.equilibrium_state_counts_full = _np.zeros((self.nthermo, self.nstates_full), dtype=_np.float64)

        pg = _ProgressReporter()
        stage = 'cset'
        with pg.context(stage=stage):
            self.csets, pcset = _cset.compute_csets_TRAM(
                self.connectivity, state_counts_full, count_matrices_full,
                equilibrium_state_counts=self.equilibrium_state_counts_full,
                ttrajs=ttrajs+equilibrium_ttrajs, dtrajs=dtrajs_full+equilibrium_dtrajs_full, bias_trajs=self.btrajs+self.equilibrium_btrajs,
                nn=self.nn, factor=self.connectivity_factor,
                callback=_IterationProgressIndicatorCallBack(pg, 'finding connected set', stage=stage))
            self.active_set = pcset

        # check for empty states
        for k in range(self.nthermo):
            if len(self.csets[k]) == 0:
                _warnings.warn(
                    'Thermodynamic state %d' % k \
                    + ' contains no samples after reducing to the connected set.', EmptyState)

        # deactivate samples not in the csets, states are *not* relabeled
        self.state_counts, self.count_matrices, self.dtrajs, _  = _cset.restrict_to_csets(
            self.csets,
            state_counts=state_counts_full, count_matrices=count_matrices_full,
            ttrajs=ttrajs, dtrajs=dtrajs_full)

        if self.equilibrium is not None:
            self.equilibrium_state_counts, _, self.equilibrium_dtrajs, _ =  _cset.restrict_to_csets(
                self.csets,
                state_counts=self.equilibrium_state_counts_full, ttrajs=equilibrium_ttrajs, dtrajs=equilibrium_dtrajs_full)
        else:
            self.equilibrium_state_counts = _np.zeros((self.nthermo, self.nstates_full), dtype=_np.intc) # (remember: no relabeling)
            self.equilibrium_dtrajs = []

        # self-consistency tests
        assert _np.all(self.state_counts >= _np.maximum(self.count_matrices.sum(axis=1), \
            self.count_matrices.sum(axis=2)))
        assert _np.all(_np.sum(
            [_np.bincount(d[d>=0], minlength=self.nstates_full) for d in self.dtrajs],
            axis=0) == self.state_counts.sum(axis=0))
        assert _np.all(_np.sum(
            [_np.bincount(t[d>=0], minlength=self.nthermo) for t, d in zip(ttrajs, self.dtrajs)],
            axis=0) == self.state_counts.sum(axis=1))
        if self.equilibrium is not None:
            assert _np.all(_np.sum(
                [_np.bincount(d[d >= 0], minlength=self.nstates_full) for d in self.equilibrium_dtrajs],
                axis=0) == self.equilibrium_state_counts.sum(axis=0))
            assert _np.all(_np.sum(
                [_np.bincount(t[d >= 0], minlength=self.nthermo) for t, d in zip(equilibrium_ttrajs, self.equilibrium_dtrajs)],
                axis=0) ==  self.equilibrium_state_counts.sum(axis=1))

        # check for empty states
        for k in range(self.state_counts.shape[0]):
            if self.count_matrices[k, :, :].sum() == 0 and self.equilibrium_state_counts[k, :].sum()==0:
                _warnings.warn(
                    'Thermodynamic state %d' % k \
                    + ' contains no transitions and no equilibrium data after reducing to the connected set.', EmptyState)

        if self.init == 'mbar' and self.biased_conf_energies is None:
            if self.direct_space:
                mbar = _mbar_direct
            else:
                mbar = _mbar
            stage = 'MBAR init.'
            with pg.context(stage=stage):
                self.mbar_therm_energies, self.mbar_unbiased_conf_energies, \
                    self.mbar_biased_conf_energies, _ = mbar.estimate(
                        (state_counts_full.sum(axis=1)+self.equilibrium_state_counts_full.sum(axis=1)).astype(_np.intc),
                        self.btrajs+self.equilibrium_btrajs, dtrajs_full+equilibrium_dtrajs_full,
                        maxiter=self.init_maxiter, maxerr=self.init_maxerr,
                        callback=_ConvergenceProgressIndicatorCallBack(
                            pg, stage, self.init_maxiter, self.init_maxerr),
                        n_conf_states=self.nstates_full)
            self.biased_conf_energies = self.mbar_biased_conf_energies.copy()

        # run estimator
        if self.direct_space:
            tram = _tram_direct
            trammbar = _trammbar_direct
        else:
            tram = _tram
            trammbar = _trammbar
        #import warnings
        #with warnings.catch_warnings() as cm:
        # warnings.filterwarnings('ignore', RuntimeWarning)
        stage = 'TRAM'
        with pg.context(stage=stage):
            if self.equilibrium is None:
                self.biased_conf_energies, conf_energies, self.therm_energies, self.log_lagrangian_mult, \
                    self.increments, self.loglikelihoods = tram.estimate(
                        self.count_matrices, self.state_counts, self.btrajs, self.dtrajs,
                        maxiter=self.maxiter, maxerr=self.maxerr,
                        biased_conf_energies=self.biased_conf_energies,
                        log_lagrangian_mult=self.log_lagrangian_mult,
                        save_convergence_info=self.save_convergence_info,
                        callback=_ConvergenceProgressIndicatorCallBack(
                            pg, stage, self.maxiter, self.maxerr, subcallback=self.callback),
                        N_dtram_accelerations=self.N_dtram_accelerations)
            else: # use trammbar
                self.biased_conf_energies, conf_energies, self.therm_energies, self.log_lagrangian_mult, \
                    self.increments, self.loglikelihoods = trammbar.estimate(
                        self.count_matrices, self.state_counts, self.btrajs, self.dtrajs,
                        equilibrium_therm_state_counts=self.equilibrium_state_counts.sum(axis=1).astype(_np.intc),
                        equilibrium_bias_energy_sequences=self.equilibrium_btrajs, equilibrium_state_sequences=self.equilibrium_dtrajs,
                        maxiter=self.maxiter, maxerr=self.maxerr,
                        save_convergence_info=self.save_convergence_info,
                        biased_conf_energies=self.biased_conf_energies,
                        log_lagrangian_mult=self.log_lagrangian_mult,
                        callback=_ConvergenceProgressIndicatorCallBack(
                            pg, stage, self.maxiter, self.maxerr, subcallback=self.callback),
                        N_dtram_accelerations=self.N_dtram_accelerations,
                        overcounting_factor=self.overcounting_factor)

        # compute models
        fmsms = [_np.ascontiguousarray((
            _tram.estimate_transition_matrix(
                self.log_lagrangian_mult, self.biased_conf_energies, self.count_matrices, None,
                K)[self.active_set, :])[:, self.active_set]) for K in range(self.nthermo)]

        active_sets = [_largest_connected_set(msm, directed=False) for msm in fmsms]
        fmsms = [_np.ascontiguousarray(
            (msm[lcc, :])[:, lcc]) for msm, lcc in zip(fmsms, active_sets)]

        models = []
        for i, (msm, acs) in enumerate(zip(fmsms, active_sets)):
            models.append(_ThermoMSM(
                msm, self.active_set[acs], self.nstates_full,
                pi=_np.exp(self.therm_energies[i] - self.biased_conf_energies[i, :]),
                dt_model=self.timestep_traj.get_scaled(self.lag)))

        # set model parameters to self
        self.set_model_params(
            models=models, f_therm=self.therm_energies, f=conf_energies[self.active_set].copy())

        return self

    def log_likelihood(self):
        r"""
        Returns the value of the log-likelihood of the converged TRAM estimate.
        """
        # TODO: check that we are estimated...
        return _tram.log_likelihood_lower_bound(
            self.log_lagrangian_mult, self.biased_conf_energies,
            self.count_matrices, self.btrajs, self.dtrajs, self.state_counts,
            None, None, None, None, None)

    def pointwise_free_energies(self, therm_state=None):
        r"""
        Computes the pointwise free energies :math:`-\log(\mu^k(x))` for all points x.

        :math:`\mu^k(x)` is the optimal estimate of the Boltzmann distribution
        of the k'th ensemble defined on the set of all samples.

        Parameters
        ----------
        therm_state : int or None, default=None
            Selects the thermodynamic state k for which to compute the
            pointwise free energies.
            None selects the "unbiased" state which is defined by having
            zero bias energy.

        Returns
        -------
        mu_k : list of numpy.ndarray(X_i, dtype=numpy.float64)
             list of the same layout as dtrajs (or ttrajs). mu_k[i][t]
             contains the pointwise free energy of the frame seen in
             trajectory i and time step t.
             Frames that are not in the connected sets get assiged an
             infinite pointwise free energy.
        """
        assert self.therm_energies is not None, \
            'MEMM has to be estimate()\'d before pointwise free energies can be calculated.'
        if therm_state is not None:
            assert therm_state<=self.nthermo
        mu = [_np.zeros(d.shape[0], dtype=_np.float64) for d in self.dtrajs+self.equilibrium_dtrajs]
        if self.equilibrium is None:
            _tram.get_pointwise_unbiased_free_energies(
                therm_state,
                self.log_lagrangian_mult, self.biased_conf_energies,
                self.therm_energies, self.count_matrices,
                self.btrajs, self.dtrajs,
                self.state_counts, None, None, mu)
        else:
            _trammbar.get_pointwise_unbiased_free_energies(
                therm_state,
                self.log_lagrangian_mult, self.biased_conf_energies,
                self.therm_energies, self.count_matrices,
                self.btrajs+self.equilibrium_btrajs, self.dtrajs+self.equilibrium_dtrajs,
                self.state_counts, None, None, mu,
                equilibrium_therm_state_counts=self.equilibrium_state_counts.sum(axis=1).astype(_np.intc),
                overcounting_factor=1.0/self.lag)
        return mu

    def mbar_pointwise_free_energies(self, therm_state=None):
        assert self.mbar_therm_energies is not None, \
            'MEMM has to be estimate()\'d with init=\'mbar\' before pointwise free energies can be calculated.'
        if therm_state is not None:
            assert therm_state<=self.nthermo
        mu = [_np.zeros(d.shape[0], dtype=_np.float64) for d in self.dtrajs+self.equilibrium_dtrajs]
        _mbar.get_pointwise_unbiased_free_energies(therm_state,
            _np.log(self.therm_state_counts_full + self.equilibrium_state_counts_full.sum(axis=1)).astype(_np.float64),
            self.btrajs+self.equilibrium_btrajs, self.mbar_therm_energies, None, mu)
        return mu

    def __getstate__(self):
        # do not pickle callable
        state = super(TRAM, self).__getstate__()
        state.pop('callback', None)
        return state




class Undefined(object): pass


class ThermoBase(object):

    __serialize_version = 0
    __serialize_fields = ('umbrella_centers', 'force_constants', 'temperatures', 'dt_traj')

    @property
    def umbrella_centers(self):
        """ The individual umbrella centers labelled accordingly to ttrajs.
        (only set, when estimated from umbrella data).
        """
        try:
            return self._umbrella_centers
        except AttributeError:
            return Undefined

    @umbrella_centers.setter
    def umbrella_centers(self, value):
        self._umbrella_centers = value

    @property
    def force_constants(self):
        """The individual force matrices labelled accordingly to ttrajs.
        (only set, when estimated from umbrella data).
        """
        try:
            return self._force_constants
        except AttributeError:
            return Undefined

    @force_constants.setter
    def force_constants(self, value):
        self._force_constants = value

    @property
    def temperatures(self):
        """ The individual temperatures labelled accordingly to ttrajs.
        (only set, when estimated from multi-temperature data).
        """
        try:
            return self._temperatures
        except AttributeError:
            return Undefined

    @temperatures.setter
    def temperatures(self, value):
        self._temperatures = value

    @property
    def dt_traj(self):
        return self._dt_traj

    @dt_traj.setter
    def dt_traj(self, value):
        # time step
        self._dt_traj = value
        from pyemma.util.units import TimeUnit
        self.timestep_traj = TimeUnit(self.dt_traj)



#!/usr/bin/env python

import os
from glob import glob

import numpy as np
import sys

import pyemma
import tables
pyemma.config.show_progress_bars = False
pyemma.config.use_trajectory_lengths_cache = False
#from pyemma.coordinates.estimation.covariance import Covariances

fast_folders_dir_src = '/nfs/group/ag_cmb/simulation-data/DESRES-Science2011-FastProteinFolding/'
#fast_folders_dir_dest = os.path.join(os.path.expanduser('~'), 'NO_BACKUP/data/feature_sel/')
fast_folders_dir_dest = '/group/ag_cmb/simulation-data/DESRES-Science2011-FastProteinFolding/'
test_systems = ['1FME', '2F4K', '2JOF', '2WAV', 'A3D', 'CLN025', 'GTT', 'lambda', 'NuG2', 'PRB', 'UVF',
                'NTL9-0-protein_fixed_CC']


def copy_data():
    from  distutils.dir_util import copy_tree
    for part in test_systems:
        pattern = fast_folders_dir_src + '/' + '*' + part + '*'
        dirs = glob(pattern)
        for d in dirs:
            print('copying %s ... ' % d)
            copy_tree(d, fast_folders_dir_dest)

def create_traj_top_pairs(test_system):
    # match trajectory files with topology
    patterns = fast_folders_dir_dest + '/*' + test_system + '*/*.dcd'
    trajs = glob(patterns)
    top = glob(fast_folders_dir_dest + '/*' + test_system + '*/*.pdb')[0]

    trajs_with_top = [(traj, top) for traj in trajs]
    return trajs_with_top


#### features
def shrake_ruply(featurizer):
    def featurize(traj):
        import mdtraj
        res = mdtraj.shrake_rupley(traj, probe_radius=0.14, n_sphere_points=960, mode='atom')
        return res

    featurizer.add_custom_func(featurize, dim=featurizer.topology.n_atoms)

def backbone(featurizer):
    featurizer.add_backbone_torsions()

def dist_ca(featurizer):
    featurizer.add_distances_ca()

def res_mindist(featurizer):
    featurizer.add_residue_mindist(scheme='ca')


features = (dist_ca, res_mindist, backbone, shrake_ruply)
#### end of features



from sklearn.model_selection import ParameterGrid
import itertools
stuff = [create_traj_top_pairs(test_system=t) for t in test_systems]
param_grid = {'traj_top': list(itertools.chain.from_iterable(stuff)),
              'feature': features,
              }

grid = ParameterGrid(param_grid)
# this should be equal to njobs for slurm!
print("num total jobs:", len(list(grid)))


def run(traj_top, feature):
    import os
    os.chdir('/group/ag_cmb/marscher/feature_sel')

    traj, top = traj_top
    print('top: {} \n\ntrajs: {}'.format(top, traj))

    reader = pyemma.coordinates.source(traj, top=top, chunk_size=5000)
    featurizer = reader.featurizer

    feature(featurizer)
    import h5py
    feature_name = feature.__name__
    file_name = "{traj}_{feat}.h5".format(traj=os.path.basename(traj), feat=feature_name)
    print("writing to", file_name)
    with h5py.File(name=file_name, mode='a') as f:
        dataset = f.create_dataset(feature_name, shape=(reader.n_frames_total(), reader.ndim),
                                   compression=32001)

        t = 0
        with reader.iterator(return_trajindex=False) as it:
            for chunk in it:
                dataset[t:t+len(chunk)] = chunk
                t += len(chunk)


def spawn(id):
    params = grid[id]
    print("<START> id=", id)
    print("launching for job id %s with params: %s" % (id, params))
    run(**params)
    print("<END> id=", id)


if __name__ == '__main__':

    first_arg = sys.argv[1]
    if first_arg == 'spawn':
        # this is the master, if we have no id take it from the grid

        from pssh import ParallelSSHClient, utils
        hosts = ['klotz', 'helix', 'nawab',
                 'zatar', 'kendall']
        client = ParallelSSHClient(hosts)
        max = 100
        outs = []
        for id in range(len(grid)):
            out = client.run_command("{exe} cache_features.py {id}".format(exe=sys.executable, id=id))
            outs.append(out)
        map(client.join, outs)
    else:
        spawn(first_arg)





# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest
import numpy as np
from pyemma.msm import bayesian_hidden_markov_model


class TestBHMM(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # load observations
        import pyemma.datasets
        obs = pyemma.datasets.load_2well_discrete().dtraj_T100K_dt10

        # don't print
        import bhmm
        bhmm.config.verbose = False
        # hidden states
        cls.nstates = 2
        # samples
        cls.nsamples = 100

        cls.lag = 10
        cls.bhmm = bayesian_hidden_markov_model([obs], cls.nstates, cls.lag, reversible=True, nsamples=cls.nsamples)

    def test_reversible(self):
        self._reversible(self.bhmm)

    def _reversible(self, msm):
        assert msm.is_reversible

    def test_lag(self):
        self._lag(self.bhmm)

    def _lag(self, msm):
        assert msm.lagtime == self.lag

    def test_nstates(self):
        self._nstates(self.bhmm)

    def _nstates(self, msm):
        assert msm.nstates == self.nstates

    def test_transition_matrix_samples(self):
        self._transition_matrix_samples(self.bhmm)

    def _transition_matrix_samples(self, msm):
        Psamples = msm.sample_f('transition_matrix')
        # shape
        assert np.array_equal(np.shape(Psamples), (self.nsamples, self.nstates, self.nstates))
        # consistency
        import msmtools.analysis as msmana
        for P in Psamples:
            assert msmana.is_transition_matrix(P)
            assert msmana.is_reversible(P)

    def test_transition_matrix_stats(self):
        self._transition_matrix_stats(self.bhmm)

    def _transition_matrix_stats(self, msm):
        import msmtools.analysis as msmana
        # mean
        Pmean = msm.sample_mean('transition_matrix')
        # test shape and consistency
        assert np.array_equal(Pmean.shape, (self.nstates, self.nstates))
        assert msmana.is_transition_matrix(Pmean)
        # std
        Pstd = msm.sample_std('transition_matrix')
        # test shape
        assert np.array_equal(Pstd.shape, (self.nstates, self.nstates))
        # conf
        L, R = msm.sample_conf('transition_matrix')
        # test shape
        assert np.array_equal(L.shape, (self.nstates, self.nstates))
        assert np.array_equal(R.shape, (self.nstates, self.nstates))
        # test consistency
        assert np.all(L <= Pmean)
        assert np.all(R >= Pmean)

    def test_eigenvalues_samples(self):
        self._eigenvalues_samples(self.bhmm)

    def _eigenvalues_samples(self, msm):
        samples = msm.sample_f('eigenvalues')
        # shape
        self.assertEqual(np.shape(samples), (self.nsamples, self.nstates))
        # consistency
        for ev in samples:
            assert np.isclose(ev[0], 1)
            assert np.all(ev[1:] < 1.0)

    def test_eigenvalues_stats(self):
        self._eigenvalues_stats(self.bhmm)

    def _eigenvalues_stats(self, msm, tol=1e-12):
        # mean
        mean = msm.sample_mean('eigenvalues')
        # test shape and consistency
        assert np.array_equal(mean.shape, (self.nstates,))
        assert np.isclose(mean[0], 1)
        assert np.all(mean[1:] < 1.0)
        # std
        std = msm.sample_std('eigenvalues')
        # test shape
        assert np.array_equal(std.shape, (self.nstates,))
        # conf
        L, R = msm.sample_conf('eigenvalues')
        # test shape
        assert np.array_equal(L.shape, (self.nstates,))
        assert np.array_equal(R.shape, (self.nstates,))
        # test consistency
        assert np.all(L-tol <= mean)
        assert np.all(R+tol >= mean)

    def test_eigenvectors_left_samples(self):
        self._eigenvectors_left_samples(self.bhmm)

    def _eigenvectors_left_samples(self, msm):
        samples = msm.sample_f('eigenvectors_left')
        # shape
        np.testing.assert_equal(np.shape(samples), (self.nsamples, self.nstates, self.nstates))
        # consistency
        for evec in samples:
            assert np.sign(evec[0,0]) == np.sign(evec[0,1])
            assert np.sign(evec[1,0]) != np.sign(evec[1,1])

    def test_eigenvectors_left_stats(self):
        self._eigenvectors_left_stats(self.bhmm)

    def _eigenvectors_left_stats(self, msm, tol=1e-12):
        # mean
        mean = msm.sample_mean('eigenvectors_left')
        # test shape and consistency
        assert np.array_equal(mean.shape, (self.nstates, self.nstates))
        assert np.sign(mean[0,0]) == np.sign(mean[0,1])
        assert np.sign(mean[1,0]) != np.sign(mean[1,1])
        # std
        std = msm.sample_std('eigenvectors_left')
        # test shape
        assert np.array_equal(std.shape, (self.nstates, self.nstates))
        # conf
        L, R = msm.sample_conf('eigenvectors_left')
        # test shape
        assert np.array_equal(L.shape, (self.nstates, self.nstates))
        assert np.array_equal(R.shape, (self.nstates, self.nstates))
        # test consistency
        assert np.all(L-tol <= mean)
        assert np.all(R+tol >= mean)

    def test_eigenvectors_right_samples(self):
        self._eigenvectors_right_samples(self.bhmm)

    def _eigenvectors_right_samples(self, msm):
        samples = msm.sample_f('eigenvectors_right')
        # shape
        np.testing.assert_equal(np.shape(samples), (self.nsamples, self.nstates, self.nstates))
        # consistency
        for evec in samples:
            assert np.sign(evec[0,0]) == np.sign(evec[1,0])
            assert np.sign(evec[0,1]) != np.sign(evec[1,1])

    def test_eigenvectors_right_stats(self):
        self._eigenvectors_right_stats(self.bhmm)

    def _eigenvectors_right_stats(self, msm, tol=1e-12):
        # mean
        mean = msm.sample_mean('eigenvectors_right')
        # test shape and consistency
        np.testing.assert_equal(mean.shape, (self.nstates, self.nstates))
        assert np.sign(mean[0,0]) == np.sign(mean[1,0])
        assert np.sign(mean[0,1]) != np.sign(mean[1,1])
        # std
        std = msm.sample_std('eigenvectors_right')
        # test shape
        assert np.array_equal(std.shape, (self.nstates, self.nstates))
        # conf
        L, R = msm.sample_conf('eigenvectors_right')
        # test shape
        assert np.array_equal(L.shape, (self.nstates, self.nstates))
        assert np.array_equal(R.shape, (self.nstates, self.nstates))
        # test consistency
        assert np.all(L-tol <= mean)
        assert np.all(R+tol >= mean)

    def test_stationary_distribution_samples(self):
        self._stationary_distribution_samples(self.bhmm)

    def _stationary_distribution_samples(self, msm):
        samples = msm.sample_f('stationary_distribution')
        # shape
        assert np.array_equal(np.shape(samples), (self.nsamples, self.nstates))
        # consistency
        for mu in samples:
            assert np.isclose(mu.sum(), 1.0)
            assert np.all(mu > 0.0)

    def test_stationary_distribution_stats(self):
        self._stationary_distribution_stats(self.bhmm)

    def _stationary_distribution_stats(self, msm, tol=1e-12):
        # mean
        mean = msm.sample_mean('stationary_distribution')
        # test shape and consistency
        assert np.array_equal(mean.shape, (self.nstates, ))
        assert np.isclose(mean.sum(), 1.0)
        assert np.all(mean > 0.0)
        assert np.max(np.abs(mean[0]-mean[1])) < 0.05
        # std
        std = msm.sample_std('stationary_distribution')
        # test shape
        assert np.array_equal(std.shape, (self.nstates, ))
        # conf
        L, R = msm.sample_conf('stationary_distribution')
        # test shape
        assert np.array_equal(L.shape, (self.nstates, ))
        assert np.array_equal(R.shape, (self.nstates, ))
        # test consistency
        assert np.all(L-tol <= mean)
        assert np.all(R+tol >= mean)

    def test_timescales_samples(self):
        self._timescales_samples(self.bhmm)

    def _timescales_samples(self, msm):
        samples = msm.sample_f('timescales')
        # shape
        np.testing.assert_equal(np.shape(samples), (self.nsamples, self.nstates-1))
        # consistency
        for l in samples:
            assert np.all(l > 0.0)

    def test_timescales_stats(self):
        self._timescales_stats(self.bhmm)

    def _timescales_stats(self, msm):
        # mean
        mean = msm.sample_mean('timescales')
        # test shape and consistency
        assert np.array_equal(mean.shape, (self.nstates-1, ))
        assert np.all(mean > 0.0)
        # std
        std = msm.sample_std('timescales')
        # test shape
        assert np.array_equal(std.shape, (self.nstates-1, ))
        # conf
        L, R = msm.sample_conf('timescales')
        # test shape
        assert np.array_equal(L.shape, (self.nstates-1, ))
        assert np.array_equal(R.shape, (self.nstates-1, ))
        # test consistency
        assert np.all(L <= mean)
        assert np.all(R >= mean)

    def test_lifetimes_samples(self):
        self._lifetimes_samples(self.bhmm)

    def _lifetimes_samples(self, msm):
        samples = msm.sample_f('lifetimes')
        # shape
        assert np.array_equal(np.shape(samples), (self.nsamples, self.nstates))
        # consistency
        for l in samples:
            assert np.all(l > 0.0)

    def test_lifetimes_stats(self):
        self._lifetimes_stats(self.bhmm)

    def _lifetimes_stats(self, msm):
        # mean
        mean = msm.sample_mean('lifetimes')
        # test shape and consistency
        assert np.array_equal(mean.shape, (self.nstates, ))
        assert np.all(mean > 0.0)
        # std
        std = msm.sample_std('lifetimes')
        # test shape
        assert np.array_equal(std.shape, (self.nstates, ))
        # conf
        L, R = msm.sample_conf('lifetimes')
        # test shape
        assert np.array_equal(L.shape, (self.nstates, ))
        assert np.array_equal(R.shape, (self.nstates, ))
        # test consistency
        assert np.all(L <= mean)
        assert np.all(R >= mean)

    def test_submodel_simple(self):
        # sanity check for submodel;
        # call should not alter self
        from copy import deepcopy
        # dtrj = np.random.randint(0, 2, size=100)
        # dtrj[np.random.randint(0, dtrj.shape[0], 3)] = 2
        # hard-coded due to stochastic failures
        dtrj = [1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 0, 0, 0, 1, 0, 1, 1, 0, 1, 1, 0, 0, 0,
                0, 2, 1, 0, 1, 1, 0, 1, 0, 1, 1, 1, 0, 1, 0, 1, 1, 0, 1, 1, 0, 1, 0, 0, 1, 0,
                1, 1, 0, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 2, 0, 0, 1, 1, 2, 0, 1, 1, 1,
                0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0, 0, 0, 1, 1, 0, 1, 1, 0]

        h = bayesian_hidden_markov_model(dtrj, 3, 2)
        h_original = deepcopy(h)

        hs = h.submodel_largest(mincount_connectivity=5)

        self.assertTrue(h == h_original)

        self.assertEqual(hs.timescales().shape[0], 1)
        self.assertEqual(hs.pi.shape[0], 2)
        self.assertEqual(hs.transition_matrix.shape, (2, 2))

    # TODO: these tests can be made compact because they are almost the same. can define general functions for testing
    # TODO: samples and stats, only need to implement consistency check individually.

class TestBHMMSpecialCases(unittest.TestCase):

    def test_separate_states(self):
        dtrajs = [np.array([0, 1, 1, 1, 1, 1, 0, 1, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 1]),
                  np.array([2, 2, 2, 2, 2, 2, 1, 1, 1, 1, 1, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 2]),]
        hmm_bayes = bayesian_hidden_markov_model(dtrajs, 3, lag=1, separate=[0], nsamples=100)
        # we expect zeros in all samples at the following indexes:
        pobs_zeros = [[0, 1, 2, 2, 2], [0, 0, 1, 2, 3]]
        for s in hmm_bayes.samples:
            assert np.allclose(s.observation_probabilities[pobs_zeros], 0)
        for strajs in hmm_bayes.sampled_trajs:
            assert strajs[0][0] == 2
            assert strajs[0][6] == 2

    def test_initialized_bhmm(self):
        import pyemma.datasets as d
        import pyemma.msm

        obs = d.load_2well_discrete().dtraj_T100K_dt10

        init_hmm = pyemma.msm.estimate_hidden_markov_model(obs, 2, 10)
        bay_hmm = pyemma.msm.estimators.BayesianHMSM(nstates=init_hmm.nstates, lag=init_hmm.lag,
                                                     stride=init_hmm.stride, init_hmsm=init_hmm)
        bay_hmm.estimate(obs)

        assert np.isclose(bay_hmm.stationary_distribution.sum(), 1)

        with self.assertRaises(NotImplementedError) as ctx:
            obs = np.copy(obs)
            assert obs[0] != np.min(obs)
            obs[0] = np.min(obs)
            bay_hmm.estimate(obs)
            self.assertIn('same data', ctx.exception.message)

    def test_initialized_bhmm_newstride(self):
        import pyemma.msm
        obs = np.random.randint(0, 2, size=1000)

        init_hmm = pyemma.msm.estimate_hidden_markov_model(obs, 2, 10)
        bay_hmm = pyemma.msm.estimators.BayesianHMSM(nstates=init_hmm.nstates, lag=init_hmm.lag,
                                                     stride='effective', init_hmsm=init_hmm)
        bay_hmm.estimate(obs)

        assert np.isclose(bay_hmm.stationary_distribution.sum(), 1)


if __name__ == "__main__":
    unittest.main()



import numpy


def variable_cols(X, tol=0.0, min_constant=0):
    """ Evaluates which columns are constant (0) or variable (1)

    Parameters
    ----------
    X : ndarray
        Matrix whose columns will be checked for constant or variable.
    tol : float
        Tolerance for float-matrices. When set to 0 only equal columns with
        values will be considered constant. When set to a positive value,
        columns where all elements have absolute differences to the first
        element of that column are considered constant.
    min_constant : int
        Minimal number of constant columns to resume operation. If at one
        point the number of constant columns drops below min_constant, the
        computation will stop and all columns will be assumed to be variable.
        In this case, an all-True array will be returned.

    Returns
    -------
    variable : bool-array
        Array with number of elements equal to the columns. True: column is
        variable / non-constant. False: column is constant.

    """
    if X is None:
        return None
    from pyemma._ext.variational.estimators.covar_c._covartools import (variable_cols_double,
                                                                        variable_cols_float,
                                                                        variable_cols_int,
                                                                        variable_cols_long,
                                                                        variable_cols_char)
    # prepare column array
    cols = numpy.zeros(X.shape[1], dtype=numpy.bool, order='C')

    if X.dtype == numpy.float64:
        completed = variable_cols_double(cols, X, tol, min_constant)
    elif X.dtype == numpy.float32:
        completed = variable_cols_float(cols, X, tol, min_constant)
    elif X.dtype == numpy.int32:
        completed = variable_cols_int(cols, X, 0, min_constant)
    elif X.dtype == numpy.int64:
        completed = variable_cols_long(cols, X, 0, min_constant)
    elif X.dtype == numpy.bool:
        completed = variable_cols_char(cols, X, 0, min_constant)
    else:
        raise TypeError('unsupported type of X: %s' % X.dtype)

    # if interrupted, return all ones. Otherwise return the variable columns as bool array
    if completed == 0:
        return numpy.ones_like(cols, dtype=numpy.bool)

    return cols



# This file is part of PyEMMA.
#
# Copyright (c) 2017 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest

from pyemma._base.progress import ProgressReporterMixin, ProgressReporter
from pyemma.util.contexts import settings
import pyemma


class TestProgress(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        import sys, os
        cls.old_std_err = sys.stderr
        sys.stderr = os.devnull

    @classmethod
    def tearDownClass(cls):
        pyemma.config.show_progress_bars = False
        import sys
        sys.stderr = cls.old_std_err

    def setUp(self):
        pyemma.config.show_progress_bars = True
        self.pg = ProgressReporterMixin()
        self.pg._progress_register(100, "test")

    def test_config_override(self):
        self.pg.show_progress = True
        with settings(show_progress_bars=False):
            assert self.pg.show_progress == False

    def test_config_2(self):
        self.pg.show_progress = False
        with settings(show_progress_bars=True):
            assert not self.pg.show_progress

    def test_ctx(self):
        pg = ProgressReporter()
        pg.register(100, 'test')
        pg.register(40, 'test2')
        try:
            with pg.context():
                pg.update(50, stage='test')
                raise Exception()
        except Exception:
            assert pg.num_registered == 0

    def test_ctx2(self):
        pg = ProgressReporter()
        assert pg.show_progress
        pg.register(100, stage='test')
        pg.register(40, stage='test2')
        try:
            with pg.context(stage='test'):
                pg.update(50, stage='test')
                raise Exception()
        except Exception:
            assert pg.num_registered == 1
            assert 'test2' in pg.registered_stages

    def test_ctx3(self):
        pg = ProgressReporter()
        assert pg.show_progress
        pg.register(100, stage='test')
        pg.register(40, stage='test2')
        pg.register(25, stage='test3')
        try:
            with pg.context(stage=('test', 'test3')):
                pg.update(50, stage='test')
                pg.update(2, stage='test3')
                raise Exception()
        except Exception:
            assert pg.num_registered == 1
            assert 'test2' in pg.registered_stages

    def test_ctx4(self):
        pg = ProgressReporter()
        pg.register(100, 'test')
        pg.register(40, 'test2')
        try:
            with pg.context():
                pg.update(50, stage='all')
                raise Exception()
        except Exception:
            assert pg.num_registered == 0

    def test_below_threshold(self):
        # show not raise
        pg = ProgressReporter()
        pg.register(2)
        pg.update(1)
        pg.set_description('dummy')


if __name__ == '__main__':
    unittest.main()



import numpy as np

from pyemma._base.serialization.serialization import SerializableMixIn
from pyemma.coordinates.data._base.datasource import DataSourceIterator, DataSource

__author__ = 'marscher'


class SourcesMerger(DataSource, SerializableMixIn):
    __serialize_version = 0
    """ Combines multiple data sources to stream from.

    Note that you are responsible you only join matching (meaningful) data sets. If one trajectory is for instance
    shorter than the another, the longer one will be truncated during iteration.

    Parameters
    ----------
    sources : list, tuple
        list of DataSources (Readers, StreamingTransformers etc.) to combine for streaming access.

    chunk: int or None
        chunk size to use for underlying iterators.
    """
    def __init__(self, sources, chunk=None):
        super(SourcesMerger, self).__init__(chunksize=chunk)
        self.sources = sources
        self._is_reader = True
        self._ndim = sum(s.ndim for s in sources)
        ntraj = sources[0].ntraj
        for s in sources[1:]:
            if s.ntraj != ntraj:
                raise ValueError('different amount of trajectories to merge: {}'
                                 .format([(s, s.ntraj) for s in sources]))
        self._ntraj = ntraj
        self._filenames = []
        for s in sources:
            self._filenames += s.filenames
        import itertools
        for pair in itertools.combinations((s.trajectory_lengths() for s in sources), 2):
            if np.any(pair[0] != pair[1]):
                raise ValueError("currently only implemented for matching datasets.")
        self._lengths = sources[0].trajectory_lengths()

    def _create_iterator(self, skip=0, chunk=0, stride=1, return_trajindex=True, cols=None):
        return _JoiningIterator(self, self.sources, skip=skip, chunk=chunk, stride=stride,
                                return_trajindex=return_trajindex, cols=cols)

    def __reduce__(self):
        return SourcesMerger, (self.sources, self.chunksize)


class _JoiningIterator(DataSourceIterator):

    def __init__(self, src, sources, skip=0, chunk=0, stride=1, return_trajindex=False, cols=None):
        super(_JoiningIterator, self).__init__(src, skip, chunk,
                                               stride, return_trajindex, cols)
        if not sources:
            raise ValueError('need some data sources.')
        self._iterators = [s.iterator(skip=skip, chunk=chunk, stride=stride,
                                      return_trajindex=return_trajindex, cols=cols)
                           for s in sources]

    def close(self):
        for it in self._iterators:
            it.close()

    def _next_chunk(self):
        # This method assumes that invoking next(iterator) handles file selections properly.
        # If one iterator raises stop iteration, this is propagated.
        chunks = []
        for it in self._iterators:
            if it.return_traj_index:
                itraj, X = next(it)
                assert itraj == self.current_trajindex
            else:
                X = next(it)
            chunks.append(X)

        res = np.hstack(chunks)
        return res

    @DataSourceIterator._select_file_guard
    def _select_file(self, itraj):
        for it in self._iterators:
            it._select_file(itraj)




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as _np
import warnings

from pyemma._ext.sklearn.base import _pprint
from pyemma.util.statistics import confidence_interval
from pyemma.util.reflection import call_member, getargspec_no_self

__author__ = 'noe'


class Model(object):
    """ Base class for PyEMMA models

    This class is inspired by sklearn's BaseEstimator class. However, we define parameter names not by the
    current class' __init__ but have to announce them. This allows us to also remember the parameters of model
    superclasses. This class can be mixed with pyEMMA and sklearn Estimators.

    """

    def __my_getstate__(self):
        state = {}

        inspect_classes = filter(lambda c: hasattr(c, '_get_model_param_names'), self.__class__.__mro__)
        for c in inspect_classes:
            state.update({k: getattr(self, k, None) for k in c._get_model_param_names()})

        return state

    def __my_setstate__(self, state):
        if state:
            for c in filter(lambda c: hasattr(c, '_get_model_param_names'), self.__class__.__mro__):
                # TODO: actually we would desire to pop from state, but this can't be done because of ThermoMSM (would pop pi twice)
                params_for_c = {k: state.get(k) for k in c._get_model_param_names()}
                c.set_model_params(self, **params_for_c)

    @classmethod
    def _get_model_param_names(cls):
        r"""Get parameter names for the model"""
        # fetch model parameters
        if hasattr(cls, 'set_model_params'):
            # introspect the constructor arguments to find the model parameters
            # to represent
            args, varargs, kw, default = getargspec_no_self(cls.set_model_params)
            if varargs is not None:
                raise RuntimeError("PyEMMA models should always specify their parameters in the signature"
                                   " of their set_model_params (no varargs). %s doesn't follow this convention."
                                   % (cls,))
            return args
        else:
            # No parameters known
            return []

    def set_model_params(self, **kw):
        for k in kw:
            setattr(self, k, kw[k])

    def update_model_params(self, **params):
        r"""Update given model parameter if they are set to specific values"""
        for key, value in params.items():
            if not hasattr(self, key):
                setattr(self, key, value)  # set parameter for the first time.
            elif getattr(self, key) is None:
                setattr(self, key, value)  # update because this parameter is still None.
            elif value is not None:
                setattr(self, key, value)  # only overwrite if set to a specific value (None does not overwrite).

    def get_model_params(self, deep=True):
        r"""Get parameters for this model.

        Parameters
        ----------
        deep: boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_model_param_names():
            # We need deprecation warnings to always be on in order to
            # catch deprecated param values.
            # This is set in utils/__init__.py but it gets overwritten
            # when running under python3 somehow.
            from pyemma.util.exceptions import PyEMMA_DeprecationWarning
            warnings.simplefilter("always", DeprecationWarning)
            warnings.simplefilter("always", PyEMMA_DeprecationWarning)
            try:
                with warnings.catch_warnings(record=True) as w:
                    value = getattr(self, key, None)
                if len(w) and w[0].category in(DeprecationWarning, PyEMMA_DeprecationWarning):
                    # if the parameter is deprecated, don't show it
                    continue
            finally:
                warnings.filters.pop(0)
                warnings.filters.pop(0)

            # XXX: should we rather test if instance of estimator?
            if deep and hasattr(value, 'get_params'):
                deep_items = list(value.get_params().items())
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def __repr__(self):
        class_name = self.__class__.__name__
        return '%s(%s)' % (class_name, _pprint(self.get_model_params(deep=False),
                                               offset=len(class_name),),)


class SampledModel(Model):

    def __init__(self, samples, conf=0.95):
        self.set_model_params(samples=samples, conf=conf)

    # TODO: maybe rename to parametrize in order to avoid confusion with set_params that has a different behavior?
    def set_model_params(self, samples=None, conf=0.95):
        self.update_model_params(samples=samples, conf=conf)

    @property
    def samples(self):
        return self._samples

    @samples.setter
    def samples(self, value):
        if value is not None:
            self.nsamples = len(value)
        self._samples = value

    def _check_samples_available(self):
        if self.samples is None:
            raise AttributeError('Model samples not available in '+str(self)+'. Call set_model_params with samples.')

    def sample_f(self, f, *args, **kwargs):
        r"""Evaluated method f for all samples

        Calls f(\*args, \*\*kwargs) on all samples.

        Parameters
        ----------
        f : method reference or name (str)
            Model method to be evaluated for each model sample

        args : arguments
            Non-keyword arguments to be passed to the method in each call

        kwargs : keyword-argments
            Keyword arguments to be passed to the method in each call

        Returns
        -------
        vals : list
            list of results of the method calls

        """
        self._check_samples_available()
        # TODO: can we use np.fromiter here? We would ne the same shape of every member for this!
        return [call_member(M, f, *args, **kwargs) for M in self.samples]

    def sample_mean(self, f, *args, **kwargs):
        r"""Sample mean of numerical method f over all samples

        Calls f(\*args, \*\*kwargs) on all samples and computes the mean.
        f must return a numerical value or an ndarray.

        Parameters
        ----------
        f : method reference or name (str)
            Model method to be evaluated for each model sample
        args : arguments
            Non-keyword arguments to be passed to the method in each call
        kwargs : keyword-argments
            Keyword arguments to be passed to the method in each call

        Returns
        -------
        mean : float or ndarray
            mean value or mean array

        """
        vals = self.sample_f(f, *args, **kwargs)
        return _np.mean(vals, axis=0)

    def sample_std(self, f, *args, **kwargs):
        r"""Sample standard deviation of numerical method f over all samples

        Calls f(\*args, \*\*kwargs) on all samples and computes the standard deviation.
        f must return a numerical value or an ndarray.

        Parameters
        ----------
        f : method reference or name (str)
            Model method to be evaluated for each model sample
        args : arguments
            Non-keyword arguments to be passed to the method in each call
        kwargs : keyword-argments
            Keyword arguments to be passed to the method in each call

        Returns
        -------
        std : float or ndarray
            standard deviation or array of standard deviations

        """
        vals = self.sample_f(f, *args, **kwargs)
        return _np.std(vals, axis=0)

    def sample_conf(self, f, *args, **kwargs):
        r"""Sample confidence interval of numerical method f over all samples

        Calls f(\*args, \*\*kwargs) on all samples and computes the confidence interval.
        Size of confidence interval is given in the construction of the
        SampledModel. f must return a numerical value or an ndarray.

        Parameters
        ----------
        f : method reference or name (str)
            Model method to be evaluated for each model sample
        args : arguments
            Non-keyword arguments to be passed to the method in each call
        kwargs : keyword-argments
            Keyword arguments to be passed to the method in each call

        Returns
        -------
        L : float or ndarray
            lower value or array of confidence interval
        R : float or ndarray
            upper value or array of confidence interval

        """
        vals = self.sample_f(f, *args, **kwargs)
        return confidence_interval(vals, conf=self.conf)



# This file is part of PyEMMA.
#
# Copyright (c) 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.



from pyemma._base.estimator import Estimator
from pyemma.coordinates.data import DataInMemory
from pyemma.coordinates.data._base.iterable import Iterable
from pyemma.util.exceptions import NotConvergedWarning


class StreamingEstimator(Estimator):
    r""" Base class for streamed estimation.

    It checks the input and wraps it in a Iterable, to be able to access the data
    in a streaming fashion.
    """

    def estimate(self, X, chunksize=None, **kwargs):
        # ensure the input is able to provide a stream
        if not isinstance(X, Iterable):
            from pyemma.util import types
            array_list = types.ensure_traj_list(X)
            X = DataInMemory(array_list, chunksize=chunksize)
        # Because we want to use pipelining methods like get_output, we have to set a data producer.
        self.data_producer = X
        X.chunksize = chunksize
        # run estimation
        try:
            super(StreamingEstimator, self).estimate(X, **kwargs)
        except NotConvergedWarning as ncw:
            self.logger.info(
                "Presumably finished estimation. Message: %s" % ncw)
        return self



import unittest
import numpy as np
from .. import running_moments

__author__ = 'noe'


class TestRunningMoments(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.X = np.random.rand(10000, 2)
        cls.Y = np.random.rand(10000, 2)
        cls.T = cls.X.shape[0]
        # Chunk size:
        cls.L = 1000
        # Number of chunks:
        cls.nchunks = cls.T // cls.L
        # Set a lag time for time-lagged tests:
        #cls.lag = 50
        # Weights references:
        cls.weights = np.random.rand(10000)
        # Trajectory weights:
        cls.trajweights = 3*np.random.rand(cls.nchunks)
        # bias the first part
        cls.X[:2000] += 1.0
        cls.Y[:2000] -= 1.0
        # column subsets
        cls.cols_2 = np.array([0])

        # direct calculation, moments of X and Y
        cls.w = np.shape(cls.X)[0]
        cls.wsym = 2*np.shape(cls.X)[0]
        cls.sx = cls.X.sum(axis=0)
        cls.sy = cls.Y.sum(axis=0)
        cls.Mxx = np.dot(cls.X.T, cls.X)
        cls.Mxy = np.dot(cls.X.T, cls.Y)
        cls.Myy = np.dot(cls.Y.T, cls.Y)
        cls.mx = cls.sx / float(cls.w)
        cls.my = cls.sy / float(cls.w)
        cls.X0 = cls.X - cls.mx
        cls.Y0 = cls.Y - cls.my
        cls.Mxx0 = np.dot(cls.X0.T, cls.X0)
        cls.Mxy0 = np.dot(cls.X0.T, cls.Y0)
        cls.Myy0 = np.dot(cls.Y0.T, cls.Y0)

        # direct calculation, symmetric moments
        cls.s_sym = cls.sx + cls.sy
        cls.Mxx_sym = np.dot(cls.X.T, cls.X) + np.dot(cls.Y.T, cls.Y)
        cls.Mxy_sym = np.dot(cls.X.T, cls.Y) + np.dot(cls.Y.T, cls.X)
        cls.m_sym = cls.s_sym / float(cls.wsym)
        cls.X0_sym = cls.X - cls.m_sym
        cls.Y0_sym = cls.Y - cls.m_sym
        cls.Mxx0_sym = np.dot(cls.X0_sym.T, cls.X0_sym) + np.dot(cls.Y0_sym.T, cls.Y0_sym)
        cls.Mxy0_sym = np.dot(cls.X0_sym.T, cls.Y0_sym) + np.dot(cls.Y0_sym.T, cls.X0_sym)

        # direct calculation, weighted moments:
        cls.wesum = np.sum(cls.weights)
        cls.sx_w = (cls.weights[:, None] * cls.X).sum(axis=0)
        cls.sy_w = (cls.weights[:, None] * cls.Y).sum(axis=0)
        cls.Mxx_w = np.dot((cls.weights[:, None] * cls.X).T, cls.X)
        cls.Mxy_w = np.dot((cls.weights[:, None] * cls.X).T, cls.Y)
        cls.mx_w = cls.sx_w / float(cls.wesum)
        cls.my_w = cls.sy_w / float(cls.wesum)
        cls.X0_w = cls.X - cls.mx_w
        cls.Y0_w = cls.Y - cls.my_w
        cls.Mxx0_w = np.dot((cls.weights[:, None] * cls.X0_w).T, cls.X0_w)
        cls.Mxy0_w = np.dot((cls.weights[:, None] * cls.X0_w).T, cls.Y0_w)

        # direct calculation, weighted symmetric moments
        cls.s_sym_w = cls.sx_w + cls.sy_w
        cls.Mxx_sym_w = np.dot((cls.weights[:, None] * cls.X).T, cls.X) + np.dot((cls.weights[:, None] * cls.Y).T, cls.Y)
        cls.Mxy_sym_w = np.dot((cls.weights[:, None] * cls.X).T, cls.Y) + np.dot((cls.weights[:, None] * cls.Y).T, cls.X)
        cls.m_sym_w = cls.s_sym_w / float(2 * cls.wesum)
        cls.X0_sym_w = cls.X - cls.m_sym_w
        cls.Y0_sym_w = cls.Y - cls.m_sym_w
        cls.Mxx0_sym_w = np.dot((cls.weights[:, None] *cls.X0_sym_w).T, cls.X0_sym_w) + np.dot((cls.weights[:, None] *cls.Y0_sym_w).T, cls.Y0_sym_w)
        cls.Mxy0_sym_w = np.dot((cls.weights[:, None] *cls.X0_sym_w).T, cls.Y0_sym_w) + np.dot((cls.weights[:, None] *cls.Y0_sym_w).T, cls.X0_sym_w)

        return cls

    def test_XX_withmean(self):
        # many passes
        cc = running_moments.RunningCovar(remove_mean=False)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L])
        np.testing.assert_allclose(cc.weight_XX(), self.T)
        np.testing.assert_allclose(cc.sum_X(), self.sx)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx)
        cc = running_moments.RunningCovar(remove_mean=False, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx[:, self.cols_2])
        cc = running_moments.RunningCovar(remove_mean=False, diag_only=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx))

    def test_XX_meanfree(self):
        # many passes
        cc = running_moments.RunningCovar(remove_mean=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L])
        np.testing.assert_allclose(cc.weight_XX(), self.T)
        np.testing.assert_allclose(cc.sum_X(), self.sx)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0)
        cc = running_moments.RunningCovar(remove_mean=True, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0[:, self.cols_2])
        cc = running_moments.RunningCovar(remove_mean=True, diag_only=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx0))

    def test_XXXY_withmean(self):
        # many passes
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.weight_XY(), self.T)
        np.testing.assert_allclose(cc.sum_X(), self.sx)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx)
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy)
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy[:, self.cols_2])
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, diag_only=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx))
        np.testing.assert_allclose(cc.moments_XY(), np.diag(self.Mxy))

    def test_XXXY_meanfree(self):
        # many passes
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.weight_XY(), self.T)
        np.testing.assert_allclose(cc.sum_X(), self.sx)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0)
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0)
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0[:, self.cols_2])
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, diag_only=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx0))
        np.testing.assert_allclose(cc.moments_XY(), np.diag(self.Mxy0))

    def test_XXXY_weighted_withmean(self):
        # many passes
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False)
        for i in range(0, self.T, self.L):
            iX = self.X[i:i+self.L, :]
            iY = self.Y[i:i+self.L, :]
            iwe = self.weights[i:i+self.L]
            cc.add(iX, iY, weights=iwe)
        np.testing.assert_allclose(cc.weight_XY(), self.wesum)
        np.testing.assert_allclose(cc.sum_X(), self.sx_w)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx_w)
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy_w)
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            iX = self.X[i:i+self.L, :]
            iY = self.Y[i:i+self.L, :]
            iwe = self.weights[i:i+self.L]
            cc.add(iX, iY, weights=iwe)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx_w[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy_w[:, self.cols_2])
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, diag_only=True)
        for i in range(0, self.T, self.L):
            iX = self.X[i:i+self.L, :]
            iY = self.Y[i:i+self.L, :]
            iwe = self.weights[i:i+self.L]
            cc.add(iX, iY, weights=iwe)
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx_w))
        np.testing.assert_allclose(cc.moments_XY(), np.diag(self.Mxy_w))

    def test_XXXY_weighted_meanfree(self):
        # many passes
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True)
        for i in range(0, self.T, self.L):
            iX = self.X[i:i+self.L, :]
            iY = self.Y[i:i+self.L, :]
            iwe = self.weights[i:i+self.L]
            cc.add(iX, iY, weights=iwe)
        np.testing.assert_allclose(cc.weight_XY(), self.wesum)
        np.testing.assert_allclose(cc.sum_X(), self.sx_w)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0_w)
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0_w)
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            iX = self.X[i:i+self.L, :]
            iY = self.Y[i:i+self.L, :]
            iwe = self.weights[i:i+self.L]
            cc.add(iX, iY, weights=iwe)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0_w[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0_w[:, self.cols_2])
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, diag_only=True)
        for i in range(0, self.T, self.L):
            iX = self.X[i:i+self.L, :]
            iY = self.Y[i:i+self.L, :]
            iwe = self.weights[i:i+self.L]
            cc.add(iX, iY, weights=iwe)
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx0_w))
        np.testing.assert_allclose(cc.moments_XY(), np.diag(self.Mxy0_w))

    def test_XXXY_sym_withmean(self):
        # many passes
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, symmetrize=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.weight_XY(), 2*self.T)
        np.testing.assert_allclose(cc.sum_X(), self.s_sym)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx_sym)
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy_sym)
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, symmetrize=True, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx_sym[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy_sym[:, self.cols_2])
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, symmetrize=True, diag_only=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx_sym))
        np.testing.assert_allclose(cc.moments_XY(), np.diag(self.Mxy_sym))

    def test_XXXY_sym_meanfree(self):
        # many passes
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, symmetrize=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.weight_XY(), 2*self.T)
        np.testing.assert_allclose(cc.sum_X(), self.s_sym)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0_sym)
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0_sym)
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, symmetrize=True, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0_sym[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0_sym[:, self.cols_2])
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, symmetrize=True, diag_only=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx0_sym))
        np.testing.assert_allclose(cc.moments_XY(), np.diag(self.Mxy0_sym))

    def test_XXXY_weighted_sym_withmean(self):
        # many passes
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, symmetrize=True)
        for i in range(0, self.T, self.L):
            iwe = self.weights[i:i+self.L]
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L], weights=iwe)
        np.testing.assert_allclose(cc.weight_XY(), 2 * self.wesum)
        np.testing.assert_allclose(cc.sum_X(), self.s_sym_w)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx_sym_w)
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy_sym_w)
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, symmetrize=True, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            iwe = self.weights[i:i+self.L]
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L], weights=iwe)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx_sym_w[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy_sym_w[:, self.cols_2])
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=False, symmetrize=True, diag_only=True)
        for i in range(0, self.T, self.L):
            iwe = self.weights[i:i+self.L]
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L], weights=iwe)
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx_sym_w))
        np.testing.assert_allclose(cc.moments_XY(), np.diag(self.Mxy_sym_w))

    def test_XXXY_weighted_sym_meanfree(self):
        # many passes
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, symmetrize=True)
        for i in range(0, self.T, self.L):
            iwe = self.weights[i:i+self.L]
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L], weights=iwe)
        np.testing.assert_allclose(cc.weight_XY(), 2*self.wesum)
        np.testing.assert_allclose(cc.sum_X(), self.s_sym_w)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0_sym_w)
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0_sym_w)
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, symmetrize=True, column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            iwe = self.weights[i:i+self.L]
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L], weights=iwe)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0_sym_w[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0_sym_w[:, self.cols_2])
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, remove_mean=True, symmetrize=True, diag_only=True)
        for i in range(0, self.T, self.L):
            iwe = self.weights[i:i+self.L]
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L], weights=iwe)
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx0_sym_w))
        np.testing.assert_allclose(cc.moments_XY(), np.diag(self.Mxy0_sym_w))

    def test_XXYY_meanfree(self):
        # many passes
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, compute_YY=True, remove_mean=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.weight_XY(), self.T)
        np.testing.assert_allclose(cc.sum_X(), self.sx)
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0)
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0)
        np.testing.assert_allclose(cc.moments_YY(), self.Myy0)
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, compute_YY=True, remove_mean=True,
                                          column_selection=self.cols_2)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), self.Mxx0[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_XY(), self.Mxy0[:, self.cols_2])
        np.testing.assert_allclose(cc.moments_YY(), self.Myy0[:, self.cols_2])
        cc = running_moments.RunningCovar(compute_XX=True, compute_XY=True, compute_YY=True, remove_mean=True,
                                          diag_only=True)
        for i in range(0, self.T, self.L):
            cc.add(self.X[i:i+self.L], self.Y[i:i+self.L])
        np.testing.assert_allclose(cc.moments_XX(), np.diag(self.Mxx0))
        np.testing.assert_allclose(cc.moments_XY(), np.diag(self.Mxy0))
        np.testing.assert_allclose(cc.moments_YY(), np.diag(self.Myy0))


if __name__ == "__main__":
    unittest.main()



failed = [
          10,
          11,
          13,
          14,
          15,
          16,
          17,
          19,
          2,
          2,
          20,
          21,
          22,
          221,
          222,
          227,
          23,
          236,
          237,
          4,
          5,
          6,
          7,
          8,
          9,
]

from celery import group
import rescore
tasks = group([rescore.run.s(id) for id in failed])
tasks.apply_async()



import numpy as np
import scipy.linalg as scl
import unittest
import pkg_resources

from pyemma.coordinates import covariance_lagged
from pyemma.coordinates.estimation.koopman import _KoopmanEstimator
from pyemma.coordinates import source


class TestEqCovar(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Basis set definition:
        cls.nf = 10
        cls.chi = np.zeros((20, cls.nf), dtype=float)
        for n in range(cls.nf):
            cls.chi[2*n:2*(n+1), n] = 1.0

        # Load simulations:
        f = np.load(pkg_resources.resource_filename(__name__, "data/test_data_koopman.npz"))
        trajs = [f[key] for key in f.keys()]
        cls.data = [cls.chi[traj, :] for traj in trajs]
        cls.source_obj = source(cls.data, chunksize=200)

        # Lag time:
        cls.tau = 10
        # Truncation for small eigenvalues:
        cls.epsilon = 1e-6
        # Chunksize:
        cls.chunksize = 200

        # Set constant mean for tests:
        cls.mean_constant = np.random.rand(cls.nf)

        # Generate _KoopmanEstimator:
        Kest = _KoopmanEstimator(cls.tau, epsilon=cls.epsilon)
        Kest.estimate(cls.source_obj, chunksize=cls.chunksize)
        cls.weight_object = Kest.weights

        # References for xx=True
        cls.mx = np.zeros(cls.nf)
        cls.my = np.zeros(cls.nf)
        cls.mx_c = np.zeros(cls.nf)
        cls.my_c = np.zeros(cls.nf)
        cls.Mxx = np.zeros((cls.nf, cls.nf))
        cls.Mxy = np.zeros((cls.nf, cls.nf))
        cls.Mxx_c = np.zeros((cls.nf, cls.nf))
        cls.Mxy_c = np.zeros((cls.nf, cls.nf))
        cls.Mxx_sym = np.zeros((cls.nf, cls.nf))
        cls.Mxy_sym = np.zeros((cls.nf, cls.nf))
        cls.Mxx_c_sym = np.zeros((cls.nf, cls.nf))
        cls.Mxy_c_sym = np.zeros((cls.nf, cls.nf))
        cls.Mxx0 = np.zeros((cls.nf, cls.nf))
        cls.Mxy0 = np.zeros((cls.nf, cls.nf))
        cls.Mxx0_sym = np.zeros((cls.nf, cls.nf))
        cls.Mxy0_sym = np.zeros((cls.nf, cls.nf))
        cls.wt = 0
        it = cls.source_obj.iterator(lag=cls.tau, return_trajindex=False)
        # Computations with data mean:
        for X, Y in it:
            w = cls.weight_object.weights(X)
            Xc = (X - cls.mean_constant[None, :]).copy()
            Yc = (Y - cls.mean_constant[None, :]).copy()
            cls.mx += np.sum(w[:, None] * X, axis=0)
            cls.my += np.sum(w[:, None] * Y, axis=0)
            cls.mx_c += np.sum(w[:, None] * Xc, axis=0)
            cls.my_c += np.sum(w[:, None] * Yc, axis=0)
            cls.Mxx += np.dot((w[:, None]*X).T, X)
            cls.Mxy += np.dot((w[:, None]*X).T, Y)
            cls.Mxx_c += np.dot((w[:, None] * Xc).T, Xc)
            cls.Mxy_c += np.dot((w[:, None] * Xc).T, Yc)
            cls.Mxx_sym += np.dot((w[:, None]*X).T, X) + np.dot((w[:, None]*Y).T, Y)
            cls.Mxy_sym += np.dot((w[:, None]*X).T, Y) + np.dot((w[:, None]*Y).T, X)
            cls.Mxx_c_sym += np.dot((w[:, None]*Xc).T, Xc) + np.dot((w[:, None]*Yc).T, Yc)
            cls.Mxy_c_sym += np.dot((w[:, None]*Xc).T, Yc) + np.dot((w[:, None]*Yc).T, Xc)
            cls.wt += w.sum()
        cls.mx /= cls.wt
        cls.my /= cls.wt
        cls.msym = 0.5*(cls.mx + cls.my)
        cls.mx_c /= cls.wt
        cls.my_c /= cls.wt
        cls.msym_c = 0.5*(cls.mx_c + cls.my_c)
        cls.Mxx /= cls.wt
        cls.Mxy /= cls.wt
        cls.Mxx_c /= cls.wt
        cls.Mxy_c /= cls.wt
        cls.Mxx_sym /= 2*cls.wt
        cls.Mxy_sym /= 2*cls.wt
        cls.Mxx_c_sym /= 2*cls.wt
        cls.Mxy_c_sym /= 2*cls.wt

        # Computations without data mean:
        it = cls.source_obj.iterator(lag=cls.tau, return_trajindex=False)
        for X, Y in it:
            w = cls.weight_object.weights(X)
            X0 = (X - cls.mx[None, :]).copy()
            Y0 = (Y - cls.my[None, :]).copy()
            X0_sym = (X - cls.msym[None, :]).copy()
            Y0_sym = (Y - cls.msym[None, :]).copy()
            cls.Mxx0 += np.dot((w[:, None]*X0).T, X0)
            cls.Mxy0 += np.dot((w[:, None]*X0).T, Y0)
            cls.Mxx0_sym += np.dot((w[:, None]*X0_sym).T, X0_sym) + np.dot((w[:, None]*Y0_sym).T, Y0_sym)
            cls.Mxy0_sym += np.dot((w[:, None]*X0_sym).T, Y0_sym) + np.dot((w[:, None]*Y0_sym).T, X0_sym)
        cls.Mxx0 /= cls.wt
        cls.Mxy0 /= cls.wt
        cls.Mxx0_sym /= 2*cls.wt
        cls.Mxy0_sym /= 2*cls.wt

    def test_XX(self):
        cc = covariance_lagged(data=self.data, c0t=False, lag=self.tau, bessel=False, weights="koopman")
        cc1 = covariance_lagged(data=self.data, c0t=False, lag=self.tau, bessel=False, weights=self.weight_object)
        assert np.allclose(cc.mean, self.mx)
        assert np.allclose(cc.C00_, self.Mxx)
        assert np.allclose(cc1.mean, self.mx)
        assert np.allclose(cc1.C00_, self.Mxx)

    def test_XX_removeconstantmean(self):
        cc = covariance_lagged(data=self.data, c0t=False, lag=self.tau, remove_constant_mean=self.mean_constant,
                               bessel=False, weights="koopman")
        cc1 = covariance_lagged(data=self.data, c0t=False, lag=self.tau, remove_constant_mean=self.mean_constant,
                                bessel=False, weights=self.weight_object)
        assert np.allclose(cc.mean, self.mx_c)
        assert np.allclose(cc.C00_, self.Mxx_c)
        assert np.allclose(cc1.mean, self.mx_c)
        assert np.allclose(cc1.C00_, self.Mxx_c)

    def test_XX_removedatamean(self):
        cc = covariance_lagged(data=self.data, c0t=False, lag=self.tau, remove_data_mean=True, bessel=False,
                               weights="koopman")
        cc1 = covariance_lagged(data=self.data, c0t=False, lag=self.tau, remove_data_mean=True, bessel=False,
                                weights=self.weight_object)
        assert np.allclose(cc.mean, self.mx)
        assert np.allclose(cc.C00_, self.Mxx0)
        assert np.allclose(cc1.mean, self.mx)
        assert np.allclose(cc1.C00_, self.Mxx0)

    def test_XY(self):
        cc = covariance_lagged(data=self.data, lag=self.tau, c0t=True, bessel=False, weights="koopman")
        cc1 = covariance_lagged(data=self.data, lag=self.tau, c0t=True, bessel=False, weights=self.weight_object)
        assert np.allclose(cc.mean, self.mx)
        assert np.allclose(cc.mean_tau, self.my)
        assert np.allclose(cc.C00_, self.Mxx)
        assert np.allclose(cc.C0t_, self.Mxy)
        assert np.allclose(cc1.mean, self.mx)
        assert np.allclose(cc1.mean_tau, self.my)
        assert np.allclose(cc1.C00_, self.Mxx)
        assert np.allclose(cc1.C0t_, self.Mxy)


    def test_XY_removeconstantmean(self):
        cc = covariance_lagged(data=self.data, lag=self.tau, c0t=True, remove_constant_mean=self.mean_constant,
                               bessel=False, weights="koopman")
        cc1 = covariance_lagged(data=self.data, lag=self.tau, c0t=True, remove_constant_mean=self.mean_constant,
                                bessel=False, weights="koopman")
        assert np.allclose(cc.mean, self.mx_c)
        assert np.allclose(cc.mean_tau, self.my_c)
        assert np.allclose(cc.C00_, self.Mxx_c)
        assert np.allclose(cc.C0t_, self.Mxy_c)
        assert np.allclose(cc1.mean, self.mx_c)
        assert np.allclose(cc1.mean_tau, self.my_c)
        assert np.allclose(cc1.C00_, self.Mxx_c)
        assert np.allclose(cc1.C0t_, self.Mxy_c)

    def test_XY_removedatamean(self):
        cc = covariance_lagged(data=self.data, lag=self.tau, c0t=True, remove_data_mean=True, bessel=False,
                               weights="koopman")
        cc1 = covariance_lagged(data=self.data, lag=self.tau, c0t=True, remove_data_mean=True, bessel=False,
                                weights=self.weight_object)
        assert np.allclose(cc.mean, self.mx)
        assert np.allclose(cc.mean_tau, self.my)
        assert np.allclose(cc.C00_, self.Mxx0)
        assert np.allclose(cc.C0t_, self.Mxy0)
        assert np.allclose(cc1.mean, self.mx)
        assert np.allclose(cc1.mean_tau, self.my)
        assert np.allclose(cc1.C00_, self.Mxx0)
        assert np.allclose(cc1.C0t_, self.Mxy0)

    def test_XY_sym(self):
        cc = covariance_lagged(data=self.data, lag=self.tau, c0t=True, reversible=True, bessel=False,
                               weights="koopman")
        cc1 = covariance_lagged(data=self.data, lag=self.tau, c0t=True, reversible=True, bessel=False,
                                weights=self.weight_object)
        assert np.allclose(cc.mean, self.msym)
        assert np.allclose(cc.C00_, self.Mxx_sym)
        assert np.allclose(cc.C0t_, self.Mxy_sym)
        assert np.allclose(cc1.mean, self.msym)
        assert np.allclose(cc1.C00_, self.Mxx_sym)
        assert np.allclose(cc1.C0t_, self.Mxy_sym)

    def test_XY_sym_removeconstantmean(self):
        cc = covariance_lagged(data=self.data, lag=self.tau, c0t=True, reversible=True,
                               remove_constant_mean=self.mean_constant, bessel=False, weights="koopman")
        cc1 = covariance_lagged(data=self.data, lag=self.tau, c0t=True, reversible=True,
                                remove_constant_mean=self.mean_constant, bessel=False, weights=self.weight_object)
        assert np.allclose(cc.mean, self.msym_c)
        assert np.allclose(cc.C00_, self.Mxx_c_sym)
        assert np.allclose(cc.C0t_, self.Mxy_c_sym)
        assert np.allclose(cc1.mean, self.msym_c)
        assert np.allclose(cc1.C00_, self.Mxx_c_sym)
        assert np.allclose(cc1.C0t_, self.Mxy_c_sym)

    def test_XY_sym_removedatamean(self):
        cc = covariance_lagged(data=self.data, lag=self.tau, c0t=True, reversible=True, remove_data_mean=True,
                               bessel=False, weights="koopman")
        cc1 = covariance_lagged(data=self.data, lag=self.tau, c0t=True, reversible=True, remove_data_mean=True,
                                bessel=False, weights=self.weight_object)
        assert np.allclose(cc.mean, self.msym)
        assert np.allclose(cc.C00_, self.Mxx0_sym)
        assert np.allclose(cc.C0t_, self.Mxy0_sym)
        assert np.allclose(cc1.mean, self.msym)
        assert np.allclose(cc1.C00_, self.Mxx0_sym)
        assert np.allclose(cc1.C0t_, self.Mxy0_sym)


if __name__ == "__main__":
    unittest.main()



import sys
sys.path.append('/home/mi/marscher/software/pycharm-2017.2/debug-eggs/pycharm-debug-py3k.egg')

import pydevd

pydevd.settrace('localhost', port=35352, stdoutToServer=True, stderrToServer=True)



import numpy as np
import re
import sys
import os

sys.path.append(os.path.expanduser('~/workspace/pyemma/feature_sel/analysis/'))
print(sys.path[-1])


def estimate_covs(test_system, feature, lag, n_covs=30, fixed_seed=False, reader=None):
    """
    Parameters
    ----------
    test_system
    feature
    files
    lag

    Returns
    -------
    estimated covariances instance.
    """
    if reader is None:
        import os
        print('current path: %s' % os.getcwd())
        import paths as p

        if feature == 'xyz':
            reader = p.create_cartesian_reader(test_system)
        else:
            reader = p.create_fragmented_reader(test_system, feature)

    assert np.all(reader.trajectory_lengths() > 0)
    assert reader.chunksize > 0
    from pyemma.coordinates.estimation.covariance import Covariances
    cov = Covariances(n_covs=n_covs, tau=lag, mode='sliding', n_save=1, fixed_seed=fixed_seed)
    cov.estimate(reader)
    return cov


def score_and_save(cov, parameters, fname, n_splits=10, k=10,
                   scoring_method='VAMP2', splitter='kfold', fixed_seed=False):
    if splitter == 'kfold':
        from sklearn.model_selection import KFold
        splitter_input = KFold(n_splits=n_splits, random_state=fixed_seed)
    elif splitter == 'shuffle':
        splitter_input = 'shuffle'
    elif splitter == 'LOO':
        from sklearn.model_selection import LeaveOneOut
        splitter_input = LeaveOneOut()
    else:
        raise ValueError('unknown splitter %s' % splitter)
    scores = np.array(cov.score_cv(n=n_splits, k=k, n_jobs=1,
                                           scoring_method=scoring_method, splitter=splitter_input,
                                           return_singular_values=False))

    print('scores:', scores.mean(), scores.std())

    np.savez_compressed(fname, parameters=parameters, scores=scores, singular_values=None)




# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''
Created on 15.02.2016

@author: marscher
'''
from pyemma._base.serialization.serialization import SerializableMixIn


class Feature(SerializableMixIn):
    __serialize_version = 0
    __serialize_fields = ('dimension', 'top')

    @property
    def dimension(self):
        return self._dim

    @dimension.setter
    def dimension(self, val):
        self._dim = int(val)

    @property
    def top(self):
        return self._top

    @top.setter
    def top(self, value):
        self._top = value

    def describe(self):
        raise NotImplementedError()

    def __eq__(self, other):
        if not isinstance(other, Feature):
            return False
        # TODO: here it should be fine to use simply the (quicker) hash func to ensure we have the same topology
        # for safety reasons, we use the much slower equality check.
        return self.dimension == other.dimension and self.top == other.top

    def __repr__(self):
        return str(self.describe())




# This file is part of PyEMMA.
#
# Copyright (c) 2016 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as _np
from pyemma.thermo import WHAM as _WHAM
from pyemma.thermo import MBAR as _MBAR
from pyemma.thermo import DTRAM as _DTRAM
from pyemma.thermo import TRAM as _TRAM
from pyemma.msm import MSM as _MSM

__all__ = [
    'plot_increments',
    'plot_loglikelihoods',
    'plot_convergence_info',
    'plot_memm_implied_timescales']


def get_estimator_label(thermo_estimator):
    if isinstance(thermo_estimator, _WHAM):
        return "WHAM"
    if isinstance(thermo_estimator, _MBAR):
        return "MBAR"
    elif isinstance(thermo_estimator, _DTRAM):
        return "dTRAM, lag=%d" % thermo_estimator.lag
    elif isinstance(thermo_estimator, _TRAM):
        return "TRAM, lag=%d" % thermo_estimator.lag
    return None


def plot_increments(thermo_estimator, ax=None):
    # TODO: write docstring
    import matplotlib.pyplot as _plt
    if ax is None:
        fig, ax = _plt.subplots()
    else:
        fig = None
    obj_list = thermo_estimator
    if not isinstance(obj_list, (list, tuple)):
        obj_list = [obj_list]
    for obj in obj_list:
        ax.plot(
            (_np.arange(obj.increments.shape[0]) + 1) * obj.save_convergence_info,
            obj.increments, '-s', label=get_estimator_label(obj))
    ax.set_xlabel(r"iteration")
    ax.set_ylabel(r"increment / kT")
    ax.semilogx()
    ax.semilogy()
    ax.legend(loc=1, fancybox=True, framealpha=0.5)
    return ax


def plot_loglikelihoods(thermo_estimator, ax=None):
    # TODO: write docstring
    import matplotlib.pyplot as _plt
    if ax is None:
        fig, ax = _plt.subplots()
    else:
        fig = None
    obj_list = thermo_estimator
    if not isinstance(obj_list, (list, tuple)):
        obj_list = [obj_list]
    for obj in obj_list:
        ax.plot(
            (_np.arange(1, obj.loglikelihoods.shape[0]) + 1) * obj.save_convergence_info,
            obj.loglikelihoods[1:] - obj.loglikelihoods[:-1], '-s', label=get_estimator_label(obj))
    ax.set_xlabel(r"iteration")
    ax.set_ylabel(r"loglikelihood increase")
    ax.semilogx()
    ax.semilogy()
    ax.legend(loc=1, fancybox=True, framealpha=0.5)
    return ax


def plot_convergence_info(thermo_estimator, axes=None):
    # TODO: write docstring
    import matplotlib.pyplot as _plt
    if axes is None:
        fs = _plt.rcParams['figure.figsize']
        fig, axes = _plt.subplots(2, 1, figsize=(fs[0], fs[1] * 1.5), sharex=True)
    else:
        assert len(axes) == 2
        fig = None
    plot_increments(thermo_estimator, ax=axes[0])
    plot_loglikelihoods(thermo_estimator, ax=axes[1])
    if fig is not None:
        fig.tight_layout()
        axes[0].set_xlabel('')
    return axes


def plot_memm_implied_timescales(thermo_estimators,
    ax=None, nits=None, therm_state=None, xlog=False, ylog=True, units='steps', dt=1.0, refs=None,
    annotate=True, **kwargs):
    import matplotlib.pyplot as _plt
    colors = ['blue', 'red', 'green', 'cyan', 'purple', 'orange', 'violet']
    # Check units and dt for user error.
    if isinstance(units, list) and len(units) != 2:
        raise TypeError("If units is a list, len(units) has to be = 2")
    if isinstance(dt, list) and len(dt) != 2:
        raise TypeError("If dt is a list, len(dt) has to be = 2")
    # Create list of units and dts for different axis
    if isinstance(units, str):
        units = [units] * 2
    if isinstance(dt, (float, int)):
        dt = [dt] * 2
    # Create figure/ax
    if ax is None:
        fig, ax = _plt.subplots()
    else:
        fig = None
    # Get timescales from all MEMM instances for the requested thermodynamic state
    lags = []
    ts = []
    for memm in thermo_estimators:
        assert isinstance(memm, (_DTRAM, _TRAM)), 'only dTRAM + TRAM accepted'
        if therm_state is None:
            assert memm.msm is not None, 'unbiased observations required'
            msm = memm.msm
        else:
            assert 0 <= therm_state < memm.nthermo, 'therm_state out of range'
            msm = memm.models[therm_state]
        assert isinstance(msm, _MSM), 'kinetic model must be MSM'
        lags.append(memm.lag)
        ts.append(msm.timescales(k=nits))
    lags = _np.asarray(lags)
    ts = _np.asarray(ts)
    srt = _np.argsort(lags)
    # Plot the implied timescales
    for i in range(ts.shape[1]):
        ax.plot(lags[srt], ts[srt, i], color=colors[i % len(colors)], **kwargs)
    # Set boundaries
    ax.set_xlim([lags.min() * dt[0], lags.max() * dt[0]])
    # Plot cutoff
    ax.plot(lags[srt] * dt[0], lags[srt] * dt[1], linewidth=2, color='black')
    ax.fill_between(
        lags[srt] * dt[0], _np.max([1, ax.get_ylim()[0]]) * _np.ones(len(lags)) * dt[1],
        lags[srt] * dt[1], alpha=0.5, color='grey')
    # formatting
    ax.set_xlabel('lag time / %s'%units[0])
    ax.set_ylabel('timescale / %s'%units[1])
    # Make the plot look pretty
    if xlog:
        ax.semilogx()
    if ylog:
        ax.semilogy()
    if therm_state is None:
        state_label = "unbiased"
    else:
        state_label = "therm_state=%d" % therm_state
    if annotate:
        ax.text(0.02, 0.95, state_label, transform=ax.transAxes)
    return ax



import os
import pyemma
from celery import Celery

from pyemma._ext.sklearn.parameter_search import ParameterGrid

print('pyemma path:', pyemma.__path__)
import numpy as np
import sys

sys.path.append(os.path.expanduser('~/feature_sel/analysis/'))
print(sys.path[-1])

features = ( 'xyz',
            #  'backbone',
            #  'side_sidechain_torsions',
            #  'dist_ca',
            #  'shrake_ruply',
            # 'res_mindist',
            # 'res_mindist_d1',
            # 'res_mindist_d2',
            # 'res_mindist_expd',
            # 'res_mindist_c_0.4',
            # 'res_mindist_c_0.5',
            # 'res_mindist_c_0.6',
            # 'res_mindist_c_0.8',
            # 'res_mindist_c_1.0',
            )

scores = (
    #'VAMPE',
    #'VAMP1',
    'VAMP2',)

grid = ParameterGrid({
    'feature': features,
    'score': scores})

print('grid len:', len(grid))
for i, e in enumerate(grid):
    print(i, e)

output_path = '/home/marscher/NO_BACKUP/1FME_joined'
#output_path = '/group/ag_cmb/marscher/feature_sel/joined3/'
broker_url = 'amqp://marscher:foobar@wallaby:5672/myvhost'
app = Celery('tasks', broker=broker_url)


def xyz(featurizer, reference):
    # create featurizer with alignment to first frame of dataset as reference
    def center(traj):
        xyz = traj.superpose(reference).xyz
        return xyz.reshape(len(xyz), -1)

    featurizer.add_custom_func(center, dim=reference.n_atoms * 3)


#### features
def shrake_ruply(featurizer):
    def featurize(traj):
        import mdtraj
        res = mdtraj.shrake_rupley(traj, probe_radius=0.14, n_sphere_points=960, mode='atom')
        return res

    featurizer.add_custom_func(featurize, dim=featurizer.topology.n_atoms)


def backbone(featurizer):
    featurizer.add_backbone_torsions(cossin=True)


def dist_ca(featurizer):
    featurizer.add_distances_ca()


def res_mindist(featurizer):
    featurizer.add_residue_mindist(scheme='closest-heavy')


def side_sidechain_torsions(featurizer):
    print("adding feature sidechain torssions to featurizer")
    from mdtraj.geometry.dihedral import indices_chi1, indices_chi2, indices_chi3, indices_chi4, indices_chi5, \
        indices_omega
    top = featurizer.topology
    indices = np.vstack((indices_chi1(top),
                         indices_chi2(top),
                         indices_chi3(top),
                         indices_chi4(top),
                         indices_chi5(top),
                         indices_omega(top)))
    assert indices.shape[1] == 4
    from mdtraj import compute_dihedrals

    def compute_side_chains(traj):
        res = compute_dihedrals(traj, indices)
        # cossin
        rad = np.dstack((np.cos(res), np.sin(res)))
        rad = rad.reshape(rad.shape[0], rad.shape[1] * rad.shape[2])
        print('shape chunk:', rad.shape)
        return rad

    featurizer.add_custom_func(compute_side_chains, dim=len(indices) * 2)


@app.task
def run(id):
    print('job number: ', id)
    params = grid[id]
    feature = params['feature']
    test_system = '1FME'
    score = params['score']
    lag = 500
    fname = os.path.join(output_path,'scores_test_sys_{test_system}_lag_{lag}_feat_{feature}_score_{scoring_method}.npz'.format(
        test_system=test_system,
        lag=lag,
        mode='sliding',
        splitter='kfold',
        scoring_method=score,
        feature=feature
    ))
    print(fname)
    if os.path.exists(fname):
        print("results file %s already exists. Skipping" % fname)
        return

    try:
        from pyemma.coordinates.estimation.covariance import Covariances
        reader = pyemma.coordinates.source([
            '/home/marscher/NO_BACKUP/1FME_joined/1FME-0.dcd',
            '/home/marscher/NO_BACKUP/1FME_joined/1FME-1.dcd',
#'/group/ag_cmb/simulation-data/DESRES-Science2011-FastProteinFolding/1FME_joined/1FME-0.dcd',
#'/group/ag_cmb/simulation-data/DESRES-Science2011-FastProteinFolding/1FME_joined/1FME-1.dcd',
            #'/srv/public/1FME_joined/1FME-0.dcd',
            #'/srv/public/1FME_joined/1FME-1.dcd',
        ],
            #top='/srv/public/1FME_joined/1FME-0-protein.pdb'
            top='/home/marscher/NO_BACKUP/1FME_joined/1FME-0-protein.pdb'
            #top='/group/ag_cmb/simulation-data/DESRES-Science2011-FastProteinFolding/1FME_joined/1FME-0-protein.pdb'
        )
        featurizer = reader.featurizer

        if feature == 'xyz':
            import mdtraj
            reference = mdtraj.load_frame('/data/scratch/marscher/1FME_joined/reference.dcd',
                                          top='/data/scratch/marscher/1FME_joined/1FME-0-protein.pdb',
                                          index=0)

            xyz(featurizer=featurizer, reference=reference)
        elif feature == 'resmin_dist':
            res_mindist(featurizer)
        elif feature == 'dist_ca':
            dist_ca(featurizer)
        elif feature == 'side_chain_torsions':
            side_sidechain_torsions(featurizer)
        elif feature == 'shrake_ruply':
            shrake_ruply(featurizer)
        elif feature.startswith('res_mindist'):
            # transform feature if needed (distances).
            res_mindist(featurizer)
            if feature.startswith('res_mindist_'):
                # if we compute contacts or transformations, add this computation.
                import re
                match = re.match('res_mindist_c_(.*)', feature)
                if match is not None:
                    arg = {'c': float(match.group(1))}
                else:
                    arg = {'scheme': re.match('res_mindist_(.*)', feature).group(1)}
                import estimate as e
                trans = e.DistTrans(dim=reader.ndim, **arg)
                trans.data_producer = reader
                reader = trans

        reader.chunksize = 10000
        cov = Covariances(n_covs=30, tau=lag, mode='sliding', n_save=2, fixed_seed=42)
        print('estimating....')
        cov.estimate(reader)
        print('finished')

        parameters = np.array({'lag': lag,
                               'splitter': 'kfold',
                               'mode': 'sliding',
                               'test_system': test_system,
                               'scoring_method': 'vamp2',
                               })

        import estimate as e
        e.score_and_save(cov, fname=fname, splitter='kfold', parameters=parameters,
                         n_splits=10, scoring_method='VAMP2')

        np.savez_compressed(fname, **results)
    except BaseException as e:
        print('bad:', e, id)
        import traceback

        traceback.print_exc(file=sys.stdout)
    else:
        print('successful')


if __name__ == '__main__':
    try:
        id_ = int(sys.argv[1])
        run(id_)
    except (IndexError, ValueError):
        print('provide an id as argument')
        sys.exit(1)



# This file is part of PyEMMA.
#
# Copyright (c) 2015, 2014 Computational Molecular Biology Group, Freie Universitaet Berlin (GER)
#
# PyEMMA is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools

import numpy as np

from pyemma._base.serialization.serialization import SerializableMixIn
from pyemma.coordinates.data._base.datasource import DataSourceIterator, DataSource, EncapsulatedIterator
from pyemma.coordinates.data.util.reader_utils import preallocate_empty_trajectory
from pyemma.util.annotators import fix_docs

# NEVER AGAIN TOUCH THIS BEAST!!!!!11111elevenoneoneone
class _FragmentedTrajectoryIterator(object):
    def __init__(self, fragmented_reader, readers, chunksize, stride, skip):
        # global time variable
        self._t = 0
        # time variable on a per-reader basis
        self._reader_t = 0
        self._readers = readers
        self._chunksize = chunksize
        self._stride = stride
        self._skip = skip
        self._frag_reader = fragmented_reader
        # lengths array per reader
        self._reader_lengths = [reader.trajectory_length(0, 1) for reader in self._readers]
        # composite trajectory length
        self._lengths = [sum(self._reader_lengths)]
        # mapping reader_index -> cumulative length
        self._cumulative_lengths = np.cumsum(self._reader_lengths)
        # current reader index
        self._reader_at = 0
        self._done = False
        self._ra_indices = None

    @property
    def ra_indices(self):
        return self._ra_indices

    @ra_indices.setter
    def ra_indices(self, value):
        self._ra_indices = value

    def __iter__(self):
        return self

    def __next__(self):
        skip = self._skip if self._t == 0 else 0
        if self._chunksize == 0:
            X = self._read_full(skip)
            return X
        else:
            # first chunk
            if self._reader_at == 0 and self._t == 0:
                self._reader_overlap = skip
                if self.ra_indices is not None:
                    self._fragment_indices = self.__get_ra_index_indices()
                    self._reader_it = self._select_next_ra_iterator()
                else:
                    self._reader_it = self._readers[self._reader_at].iterator(self._stride, return_trajindex=False,
                                                                              skip=self._reader_overlap)
            # set original chunksize
            self._reader_it.chunksize = self._chunksize
            # chunk is contained in current reader
            if self.__chunk_contained_in_current_reader():
                X = next(self._reader_it)
                L = len(X)
                self._t += L
                self._reader_t += L
                return X
            # chunk has to be collected from subsequent readers
            else:
                ndim = self._readers[0].ndim
                expected_length = self.__get_chunk_expected_length()
                X = self._allocate_chunk(expected_length, ndim)
                read = 0
                while read < expected_length or expected_length == 0:
                    # reader has data left:
                    reader_trajlen = self._reader_it.trajectory_length()
                    if reader_trajlen - self._reader_t > 0:
                        chunk = next(self._reader_it)
                        L = len(chunk)
                        X[read:read + L, :] = chunk[:]
                        read += L
                        self._reader_t += L
                    # need new reader
                    if read < expected_length or expected_length == 0:
                        self._reader_at += 1
                        self._reader_t = 0
                        if self._reader_at >= len(self._readers):
                            raise StopIteration('fragmented: out of readers')
                        self._reader_it.close()
                        if self.ra_indices is not None:
                            self._reader_it = self._select_next_ra_iterator()
                        else:
                            self._reader_overlap = self._calculate_new_overlap(self._stride,
                                                                               self._reader_lengths[self._reader_at - 1],
                                                                               self._reader_overlap)
                            self._reader_it = self._readers[self._reader_at].iterator(self._stride, skip=self._reader_overlap,
                                                                                      return_trajindex=False)
                        if expected_length - read > 0:
                            self._reader_it.chunksize = expected_length - read
                self._t += read
                return X

    next = __next__

    def _select_next_ra_iterator(self):
        assert self._reader_at < len(self._readers)
        ra_indices = self.__get_ifrag_ra_indices(self._fragment_indices, self._reader_at)
        while len(ra_indices) == 0:
            self._reader_at += 1
            if self._reader_at < len(self._readers):
                ra_indices = self.__get_ifrag_ra_indices(self._fragment_indices, self._reader_at)
            else:
                raise RuntimeError("This should not happen as the loop is supposed to be terminated in __next__.")
        return self._readers[self._reader_at].iterator(ra_indices, return_trajindex=False)

    def __get_chunk_expected_length(self):
        if self.ra_indices is not None:
            expected_length = min(self._chunksize, len(self.ra_indices) - self._t)
        else:
            expected_length = min(self._chunksize, sum(self._traj_lengths(self._stride)) - self._t)
        return expected_length

    def __chunk_contained_in_current_reader(self):
        trajlen = self._reader_it.trajectory_length()
        return trajlen - self._reader_t - self._chunksize > 0

    def _allocate_chunk(self, expected_length, ndim):
        from pyemma.coordinates.data.feature_reader import FeatureReader
        if all(isinstance(r, FeatureReader) and r._return_traj_obj for r in self._readers):
            X = preallocate_empty_trajectory(n_frames=expected_length,
                                             top=self._readers[0].featurizer.topology)
        else:
            X = np.empty((expected_length, ndim), dtype=self._frag_reader.output_type())

        return X

    def _read_full(self, skip):
        if self._ra_indices is not None:
            fragment_indices = self.__get_ra_index_indices()
            ndim = self._readers[0].ndim
            length = len(self.ra_indices)
            X = self._allocate_chunk(length, ndim)
            L = 0
            for ifrag, r in enumerate(self._readers):
                indices = self.__get_ifrag_ra_indices(fragment_indices, ifrag)
                if len(indices) > 0:
                    ifrag_data = None
                    for ifrag_data in r.iterator(chunk=0, stride=indices, return_trajindex=False):
                        pass
                    l = len(ifrag_data)
                    X[L:L+l, :] = ifrag_data[:]
                    L += l
            return X
        else:
            overlap = skip
            self._skip = overlap
            ndim = len(np.zeros(self._readers[0].dimension())[0::])
            length = sum(self._traj_lengths(self._stride))
            X = self._allocate_chunk(length, ndim)
            for idx, r in enumerate(self._readers):
                _skip = overlap
                # if stride doesn't divide length, one has to offset the next trajectory
                overlap = self._calculate_new_overlap(self._stride, self._reader_lengths[idx], overlap)
                chunksize = min(length, r.trajectory_length(0, self._stride, skip=_skip))
                it = r.iterator(stride=self._stride, skip=_skip, chunk=chunksize, return_trajindex=True)
                with it:
                    for itraj, data in it:
                        L = len(data)
                        if L > 0:
                            X[self._t:self._t + L, :] = data[:]
                        self._t += L
            return X

    def __get_ifrag_ra_indices(self, fragment_indices, ifrag):
        assert ifrag < len(self._readers)
        offset = self._cumulative_lengths[ifrag - 1] if ifrag > 0 else 0
        frag_inds = fragment_indices[ifrag]
        ra = self.ra_indices[tuple(frag_inds)] - offset
        indices = np.zeros((len(ra), 2), dtype=int)
        indices[:, 1] = ra.squeeze()
        return indices

    def __get_ra_index_indices(self):
        """
        Returns a list containing indices of the ra_index array, which correspond to the separate trajectory fragments,
        i.e., ra_indices[fragment_indices[itraj]] are the ra indices for itraj (plus some offset by
        cumulative length)
        """
        fragment_indices = []
        for idx, cumlen in enumerate(self._cumulative_lengths):
            cumlen_prev = self._cumulative_lengths[idx - 1] if idx > 0 else 0
            fragment_indices.append([np.argwhere(
                np.logical_and(self.ra_indices >= cumlen_prev, self.ra_indices < cumlen)
            )])
        return fragment_indices

    def _traj_lengths(self, stride):
        return np.array([(l - self._skip - 1) // stride + 1 for l in self._lengths], dtype=int)

    @staticmethod
    def _calculate_new_overlap(stride, traj_len, skip):
        """
        Given two trajectories T_1 and T_2, this function calculates for the first trajectory an overlap, i.e.,
        a skip parameter for T_2 such that the trajectory fragments T_1 and T_2 appear as one under the given stride.

        Idea for deriving the formula: It is

        K = ((traj_len - skip - 1) // stride + 1) = #(data points in trajectory of length (traj_len - skip)).

        Therefore, the first point's position that is not contained in T_1 anymore is given by

        pos = skip + s * K.

        Thus the needed skip of T_2 such that the same stride parameter makes T_1 and T_2 "look as one" is

        overlap = pos - traj_len.

        :param stride: the (global) stride parameter
        :param traj_len: length of T_1
        :param skip: skip of T_1
        :return: skip of T_2
        """
        overlap = stride * ((traj_len - skip - 1) // stride + 1) - traj_len + skip
        return overlap

    def close(self):
        if hasattr(self, '_reader_it'):
            self._reader_it.close()


class FragmentIterator(EncapsulatedIterator):
    """
    outer iterator, which encapsulates _FragmentedTrajectoryIterator
    """
    @EncapsulatedIterator._select_file_guard
    def _select_file(self, itraj):
        self.close()
        self._it = _FragmentedTrajectoryIterator(self._data_source, self._data_source._readers[itraj],
                                                     self.chunksize, self.stride, self.skip)
        if not self.uniform_stride:
            self._it.ra_indices = self.ra_indices_for_traj(itraj)

    # We have to delegate the chunksize to the underlying iterator, because it is not a DataSourceIterator itself.
    # All other properties are passed during the creation during changing files.
    @EncapsulatedIterator.chunksize.setter
    def chunksize(self, value):
        super(FragmentIterator, self.__class__).chunksize.__set__(self, value)
        self._it._chunksize = value


@fix_docs
class FragmentedTrajectoryReader(DataSource, SerializableMixIn):
    __serialize_version = 0
    """ Reader for fragmented trajectory data eg. (traj0_part0.xtc, traj_0_part1.xtc, ... )

    Parameters
    ----------
    trajectories: nested list or nested tuple, 1 level depth
        grouped trajectory parts, eg.:
        [['traj1_0.xtc', 'traj1_1.xtc'], 'traj2_full.xtc'], ['traj3_0.xtc, ...]]
    topologyfile, str, default None

    chunksize: int, default 1000

    featurizer: MDFeaturizer, default None
    """

    def __init__(self, trajectories, topologyfile=None, chunksize=1000, featurizer=None):
        self._args = (trajectories, topologyfile, chunksize, featurizer)
        # sanity checks
        assert isinstance(trajectories, (list, tuple)), "input trajectories should be of list or tuple type"
        # if it contains no further list: treat as single trajectory
        if not any([isinstance(traj, (list, tuple)) for traj in trajectories]):
            trajectories = [trajectories]
        # if not list of lists, treat as single-element-fragment-trajectory
        trajectories = [traj if isinstance(traj, (list, tuple)) else [traj] for traj in trajectories]
        # some trajectory should be provided
        assert len(trajectories) > 0, "no input trajectories provided"
        # call super
        super(FragmentedTrajectoryReader, self).__init__(chunksize=chunksize)
        self._is_reader = True
        # number of trajectories
        self._ntraj = len(trajectories)
        # store readers
        from pyemma.coordinates.api import source

        self._readers = [[source(input_item, features=featurizer, top=topologyfile, chunksize=chunksize)
                          for input_item in trajectories[itraj]] for itraj in range(0, self._ntraj)]

        # check all readers have same dimension
        if not len(set(itraj_r.ndim for r in self._readers for itraj_r in r)) == 1:
            # lookup the evil reader:
            last_dim = -1
            for r in self._readers:
                for itraj_r in r:
                    if last_dim == -1:
                        last_dim = itraj_r.ndim
                    if itraj_r.ndim != last_dim:
                        raise ValueError("%s has different dimension (%i) than expected (%i)"
                                         % (itraj_r.describe(), itraj_r.ndim, last_dim))

        from collections import defaultdict
        self._reader_by_filename = defaultdict(list)
        for r in self._readers:
            for itraj_r in r:
                for filename in itraj_r.filenames:
                    self._reader_by_filename[filename].append(itraj_r)

        # lengths array per reader
        self._reader_lengths = [[reader.trajectory_length(0, 1)
                                 for reader in self._readers[itraj]] for itraj in range(0, self._ntraj)]
        # composite trajectory length
        self._lengths = [sum(self._reader_lengths[itraj]) for itraj in range(0, self._ntraj)]
        # mapping reader_index -> cumulative length
        self._cumulative_lengths = [np.cumsum(self._reader_lengths[itraj]) for itraj in range(0, self._ntraj)]
        # store trajectory files
        self._trajectories = trajectories
        self._filenames = trajectories

        # random-accessible
        #self._is_random_accessible = all(r._is_random_accessible for r in self._readers[itraj]
        #                                 for itraj in range(0, self._ntraj))

    @property
    def filenames_flat(self):
        flat_readers = itertools.chain.from_iterable(self._readers)
        return [f for flat_reader in flat_readers for f in flat_reader.filenames]

    def reader_by_filename(self, filename):
        res = self._reader_by_filename[filename]
        if isinstance(res, list) and len(res) == 1:
            res = res[0]
        return res

    def _create_iterator(self, skip=0, chunk=0, stride=1, return_trajindex=True, cols=None):
        return FragmentIterator(self, skip=skip, chunk=chunk, stride=stride, return_trajindex=return_trajindex, cols=cols)

    def describe(self):
        return "[FragmentedTrajectoryReader files=%s]" % self._trajectories

    def dimension(self):
        return self._readers[0][0].dimension()

    def _index_to_reader_index(self, index, itraj):
        """
        Accepts an index parameter in [0, sum(reader_lenghts)) and returns a tuple (reader_index, local_index),
        where the tuple (reader_index, local_index) corresponds to the global frame index of the fragmented trajectory.
        :param index: the global index
        :return: a tuple (reader_index, local_index)
        """
        prev_len = 0
        for readerIndex, length in enumerate(self._cumulative_lengths[itraj]):
            if prev_len <= index < length:
                return readerIndex, index - prev_len
            prev_len = length
        raise ValueError("Requested index %s was out of bounds [0,%s)" % (index, self._cumulative_lengths[itraj][-1]))

    def _get_traj_info(self, filename):
        # get info for a fragment from specific reader
        reader = self._reader_by_filename[filename]
        return reader._get_traj_info(filename)

    def __reduce__(self):
        # serialization by storing constructor arguments
        return FragmentedTrajectoryReader, self._args
